"""Unit test for RemoteValueClimateMode objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary
from xknx.dpt.dpt_20 import HVACControllerMode, HVACOperationMode
from xknx.exceptions import ConversionError, CouldNotParseTelegram
from xknx.remote_value import (
    RemoteValueBinaryHeatCool,
    RemoteValueBinaryOperationMode,
    RemoteValueControllerMode,
    RemoteValueOperationMode,
)
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueOperationMode:
    """Test class for RemoteValueOperationMode objects."""

    def test_to_knx_operation_mode(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(xknx)
        assert remote_value.to_knx(HVACOperationMode.COMFORT) == DPTArray((0x01,))

    def test_to_knx_controller_mode(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(xknx)
        assert remote_value.to_knx(HVACControllerMode.HEAT) == DPTArray((0x01,))

    def test_to_knx_binary(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx, operation_mode=HVACOperationMode.COMFORT
        )
        assert remote_value.to_knx(HVACOperationMode.COMFORT) == DPTBinary(True)
        assert remote_value.to_knx(HVACOperationMode.ECONOMY) == DPTBinary(False)

    def test_from_knx_binary_error(self) -> None:
        """Test from_knx function with invalid payload."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx, operation_mode=HVACOperationMode.COMFORT
        )
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTArray((0x9, 0xF)))

    def test_to_knx_heat_cool(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryHeatCool(
            xknx, controller_mode=HVACControllerMode.HEAT
        )
        assert remote_value.to_knx(HVACControllerMode.HEAT) == DPTBinary(True)
        assert remote_value.to_knx(HVACControllerMode.COOL) == DPTBinary(False)

    def test_to_knx_heat_cool_error(self) -> None:
        """Test to_knx function with wrong controller mode."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryHeatCool(
            xknx, controller_mode=HVACControllerMode.HEAT
        )
        with pytest.raises(ConversionError):
            remote_value.to_knx(HVACOperationMode.STANDBY)

    def test_from_knx_operation_mode(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(xknx)
        assert remote_value.from_knx(DPTArray((0x02,))) == HVACOperationMode.STANDBY

    def test_from_knx_controller_mode(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(xknx)
        assert (
            remote_value.from_knx(DPTArray((0x02,)))
            == HVACControllerMode.MORNING_WARMUP
        )

    def test_from_knx_binary_heat_cool(self) -> None:
        """Test from_knx function with invalid payload."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryHeatCool(
            xknx, controller_mode=HVACControllerMode.HEAT
        )
        with pytest.raises(CouldNotParseTelegram):
            remote_value.from_knx(DPTArray((0x9, 0xF)))

    def test_from_knx_binary(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx, operation_mode=HVACOperationMode.COMFORT
        )
        assert remote_value.from_knx(DPTBinary(True)) == HVACOperationMode.COMFORT
        assert remote_value.from_knx(DPTBinary(False)) is None

    def test_from_knx_heat_cool(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryHeatCool(
            xknx, controller_mode=HVACControllerMode.HEAT
        )
        assert remote_value.from_knx(DPTBinary(True)) == HVACControllerMode.HEAT
        assert remote_value.from_knx(DPTBinary(False)) == HVACControllerMode.COOL

    def test_from_knx_unsupported_operation_mode(self) -> None:
        """Test from_knx function with unsupported operation."""
        xknx = XKNX()
        with pytest.raises(ConversionError):
            RemoteValueBinaryHeatCool(xknx, controller_mode=HVACControllerMode.NODEM)

    def test_from_knx_unknown_operation_mode(self) -> None:
        """Test from_knx function with unsupported operation."""
        xknx = XKNX()
        with pytest.raises(ConversionError):
            RemoteValueBinaryHeatCool(xknx, controller_mode=None)

    def test_to_knx_error_operation_mode(self) -> None:
        """Test to_knx function with wrong parameter."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(256)
        with pytest.raises(ConversionError):
            remote_value.to_knx("256")
        with pytest.raises(ConversionError):
            remote_value.to_knx(HVACControllerMode.HEAT)

    def test_to_knx_error_controller_mode(self) -> None:
        """Test to_knx function with wrong parameter."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(256)
        with pytest.raises(ConversionError):
            remote_value.to_knx("256")
        with pytest.raises(ConversionError):
            remote_value.to_knx(HVACOperationMode.ECONOMY)

    def test_to_knx_error_binary(self) -> None:
        """Test to_knx function with wrong parameter."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx, operation_mode=HVACOperationMode.ECONOMY
        )
        with pytest.raises(ConversionError):
            remote_value.to_knx(256)
        with pytest.raises(ConversionError):
            remote_value.to_knx(True)
        with pytest.raises(ConversionError):
            remote_value.to_knx(HVACControllerMode.HEAT)

    async def test_set_operation_mode(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(
            xknx, group_address=GroupAddress("1/2/3")
        )
        remote_value.set(HVACOperationMode.ECONOMY)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x03,))),
        )
        remote_value.set(HVACOperationMode.BUILDING_PROTECTION)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x04,))),
        )

    async def test_set_controller_mode(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(
            xknx, group_address=GroupAddress("1/2/3")
        )
        remote_value.set(HVACControllerMode.COOL)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x03,))),
        )
        remote_value.set(HVACControllerMode.NIGHT_PURGE)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x04,))),
        )

    async def test_set_binary(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx,
            group_address=GroupAddress("1/2/3"),
            operation_mode=HVACOperationMode.STANDBY,
        )
        remote_value.set(HVACOperationMode.STANDBY)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(True)),
        )
        remote_value.set(HVACOperationMode.BUILDING_PROTECTION)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(False)),
        )

    def test_process_operation_mode(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(
            xknx, group_address=GroupAddress("1/2/3")
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x00,))),
        )
        remote_value.process(telegram)
        assert remote_value.value == HVACOperationMode.AUTO

    def test_process_controller_mode(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(
            xknx, group_address=GroupAddress("1/2/3")
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x00,))),
        )
        remote_value.process(telegram)
        assert remote_value.value == HVACControllerMode.AUTO

    def test_process_binary(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx,
            group_address=GroupAddress("1/2/3"),
            operation_mode=HVACOperationMode.BUILDING_PROTECTION,
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(True)),
        )
        remote_value.process(telegram)
        assert remote_value.value == HVACOperationMode.BUILDING_PROTECTION

    def test_to_process_error_operation_mode(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueOperationMode(
            xknx, group_address=GroupAddress("1/2/3")
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None

    def test_to_process_error_binary_operation_mode(self) -> None:
        """Test processing invalid payload."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryOperationMode(
            xknx,
            group_address="1/2/3",
            operation_mode=HVACOperationMode.COMFORT,
        )
        invalid_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(
                DPTArray((1,)),
            ),
        )
        assert remote_value.process(telegram=invalid_telegram) is False

    def test_to_process_error_controller_mode(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueControllerMode(
            xknx, group_address=GroupAddress("1/2/3")
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None

    def test_to_process_error_heat_cool(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueBinaryHeatCool(
            xknx,
            group_address=GroupAddress("1/2/3"),
            controller_mode=HVACControllerMode.COOL,
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01,))),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Manages a connection to the KNX bus."""

from __future__ import annotations

from enum import Enum, auto
import os
from typing import Any

from xknx.secure import Keyring
from xknx.telegram.address import IndividualAddress, IndividualAddressableType

from .const import DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT
from .gateway_scanner import GatewayScanFilter


class ConnectionType(Enum):
    """Enum class for different types of KNX/IP Connections."""

    AUTOMATIC = auto()
    ROUTING = auto()
    ROUTING_SECURE = auto()
    TUNNELING = auto()
    TUNNELING_TCP = auto()
    TUNNELING_TCP_SECURE = auto()


class ConnectionConfig:
    """
    Connection configuration.

    Handles:
    * type of connection:
        * AUTOMATIC for using GatewayScanner for searching and finding KNX/IP devices in the network.
        * ROUTING use KNX/IP multicast routing.
        * TUNNELING connect to a specific KNX/IP tunneling device via UDP.
        * TUNNELING_TCP connect to a specific KNX/IP tunneling v2 device via TCP.
    * individual address:
        * AUTOMATIC use a specific tunnel endpoint from a given knxkeys file
        * ROUTING the individual address used as source address for routing
        * TCP TUNNELING request a specific tunnel endpoint
        * SECURE TUNNELING use a specific tunnel endpoint from the knxkeys file
    * local_ip: Local ip or interface name though which xknx should connect.
    * gateway_ip: IP or hostname of KNX/IP tunneling device.
    * gateway_port: Port of KNX/IP tunneling device.
    * route_back: For UDP TUNNELING connection.
        The KNXnet/IP Server shall use the IP address and port in the received IP package
        as the target IP address or port number for the response to the KNXnet/IP Client.
    * multicast_group: Multicast group for KNXnet/IP routing.
    * multicast_port: Multicast port for KNXnet/IP routing.
    * auto_reconnect: Auto reconnect to KNX/IP tunneling device if open connection was lost.
    * auto_reconnect_wait: Delay in seconds before attempting subsequent reconnects after an initial reconnect attempt fails.
    * scan_filter: For AUTOMATIC connection, limit scan with the given filter
    * threaded: Run connection logic in separate thread to avoid concurrency issues in HA
    * secure_config: KNX Secure config to use
    """

    def __init__(
        self,
        *,
        connection_type: ConnectionType = ConnectionType.AUTOMATIC,
        individual_address: IndividualAddressableType | None = None,
        local_ip: str | None = None,
        local_port: int = 0,
        gateway_ip: str | None = None,
        gateway_port: int = DEFAULT_MCAST_PORT,
        route_back: bool = False,
        multicast_group: str = DEFAULT_MCAST_GRP,
        multicast_port: int = DEFAULT_MCAST_PORT,
        auto_reconnect: bool = True,
        auto_reconnect_wait: int = 3,
        scan_filter: GatewayScanFilter | None = None,
        threaded: bool = False,
        secure_config: SecureConfig | None = None,
    ) -> None:
        """Initialize ConnectionConfig class."""
        self.connection_type = connection_type
        self.individual_address = (
            IndividualAddress(individual_address) if individual_address else None
        )
        self.local_ip = local_ip
        self.local_port = local_port
        self.gateway_ip = gateway_ip
        self.gateway_port = gateway_port
        self.route_back = route_back
        self.multicast_group = multicast_group
        self.multicast_port = multicast_port
        self.auto_reconnect = auto_reconnect
        self.auto_reconnect_wait = auto_reconnect_wait
        self.scan_filter = scan_filter or GatewayScanFilter()
        self.threaded = threaded
        self.secure_config = secure_config

    def __eq__(self, other: object) -> bool:
        """Equality for ConnectionConfig class (used in unit tests)."""
        return self.__dict__ == other.__dict__


class SecureConfig:
    """
    Secure configuration.

    Handles:
    * backbone_key: Key used for KNX Secure Routing in hex representation.
    * latency_ms: Latency in milliseconds for KNX Secure Routing.
    * user_id: The user id to use when initializing the secure tunnel.
    * device_authentication_password: the authentication password to use when connecting to the tunnel.
    * user_password: the user password for knx secure.
    * knxkeys_file_path: Full path to the knxkeys file including the file name.
    * knxkeys_password: Password to decrypt the knxkeys file.
    * keyring: Already-loaded keyring if it shouldn't be read from the knxkeys file.
    """

    def __init__(
        self,
        *,
        backbone_key: str | None = None,
        latency_ms: int | None = None,
        user_id: int | None = None,
        device_authentication_password: str | None = None,
        user_password: str | None = None,
        knxkeys_file_path: str | os.PathLike[Any] | None = None,
        knxkeys_password: str | None = None,
        keyring: Keyring | None = None,
    ) -> None:
        """Initialize SecureConfig class."""
        self.backbone_key = bytes.fromhex(backbone_key) if backbone_key else None
        self.latency_ms = latency_ms
        self.user_id = user_id
        self.device_authentication_password = device_authentication_password
        self.user_password = user_password
        self.knxkeys_file_path = knxkeys_file_path
        self.knxkeys_password = knxkeys_password
        self.keyring = keyring

    def __eq__(self, other: object) -> bool:
        """Equality for SecureConfig class (used in unit tests)."""
        return self.__dict__ == other.__dict__



"""Example for connecting to a KNX IP Secure Tunnel."""

import asyncio
import logging

from xknx import XKNX
from xknx.io import ConnectionConfig, ConnectionType, SecureConfig
from xknx.tools import group_value_write

logging.basicConfig(level=logging.INFO)
logging.getLogger("xknx.log").level = logging.DEBUG
logging.getLogger("xknx.knx").level = logging.DEBUG


async def main() -> None:
    """Test connection with IP secure tunnelling."""

    connection_config = ConnectionConfig(
        connection_type=ConnectionType.TUNNELING_TCP_SECURE,
        gateway_ip="192.168.1.188",
        individual_address="1.0.11",
        secure_config=SecureConfig(
            knxkeys_file_path="/home/marvin/testcase.knxkeys",
            knxkeys_password="password",
        ),
    )
    xknx = XKNX(connection_config=connection_config)

    await xknx.start()
    print("Tunnel connected")
    group_value_write(xknx, "1/0/32", True)
    await asyncio.sleep(5)
    await xknx.stop()


asyncio.run(main())



"""Implementation of KNX DPT 3 4-bit control."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from xknx.exceptions import ConversionError

from .dpt import DPTComplex, DPTComplexData
from .dpt_1 import Step, UpDown
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class ControlDimming(DPTComplexData):
    """
    Class for dimming control.

    Range is subdivided into `2**(step_code-1)`; `step_code=0` indicates break.
    """

    control: Step
    step_code: int  # 1..7 higher is more intervals -> slower; 0 stop

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> ControlDimming:
        """Init from a dictionary."""
        try:
            control = Step.parse(data["control"])
            step_code = int(data["step_code"])
        except (KeyError, TypeError, ValueError) as err:
            raise ValueError(f"Invalid value for ControlDimming: {err}") from err
        return cls(control=control, step_code=step_code)

    def as_dict(self) -> dict[str, int | str]:
        """Create a JSON serializable dictionary."""
        return {
            "control": self.control.name.lower(),
            "step_code": self.step_code,
        }


class DPTControlDimming(DPTComplex[ControlDimming]):
    """
    Abstraction for KNX 4-Bit dimming control.

    DPT 3.007
    """

    data_type = ControlDimming
    payload_type = DPTBinary
    payload_length = 4
    dpt_main_number = 3
    dpt_sub_number = 7
    value_type = "control_dimming"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> ControlDimming:
        """Parse/deserialize from KNX/IP payload data."""
        raw = cls.validate_payload(payload)[0]
        return ControlDimming(
            control=Step((raw & 0b1000) >> 3),
            step_code=raw & 0b0111,
        )

    @classmethod
    def _to_knx(cls, value: ControlDimming) -> DPTBinary:
        """Serialize to KNX/IP raw data."""
        if not 0 <= value.step_code <= 7:
            raise ConversionError("Invalid value for step_code: must be 0..7")
        return DPTBinary(value.control.value << 3 | value.step_code)


@dataclass(slots=True)
class ControlBlinds(DPTComplexData):
    """
    Class for blinds control.

    Range is subdivided into `2**(step_code-1)`; `step_code=0` indicates break.
    """

    control: UpDown
    step_code: int  # 1..7 higher is more intervals -> slower; 0 stop

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> ControlBlinds:
        """Init from a dictionary."""
        try:
            control = UpDown.parse(data["control"])
            step_code = int(data["step_code"])
        except (KeyError, TypeError, ValueError) as err:
            raise ValueError(f"Invalid value for ControlBlinds: {err}") from err
        return cls(control=control, step_code=step_code)

    def as_dict(self) -> dict[str, int | str]:
        """Create a JSON serializable dictionary."""
        return {
            "control": self.control.name.lower(),
            "step_code": self.step_code,
        }


class DPTControlBlinds(DPTComplex[ControlBlinds]):
    """
    Abstraction for KNX 4-Bit dimming control.

    DPT 3.008
    """

    data_type = ControlBlinds
    payload_type = DPTBinary
    payload_length = 4
    dpt_main_number = 3
    dpt_sub_number = 8
    value_type = "control_blinds"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> ControlBlinds:
        """Parse/deserialize from KNX/IP payload data."""
        raw = cls.validate_payload(payload)[0]
        return ControlBlinds(
            control=UpDown((raw & 0b1000) >> 3),
            step_code=raw & 0b0111,
        )

    @classmethod
    def _to_knx(cls, value: ControlBlinds) -> DPTBinary:
        """Serialize to KNX/IP raw data."""
        if not 0 <= value.step_code <= 7:
            raise ConversionError("Invalid value for step_code: must be 0..7")
        return DPTBinary(value.control.value << 3 | value.step_code)



"""Unit test for BinarySensor objects."""

from unittest.mock import Mock, patch

from xknx import XKNX
from xknx.devices import BinarySensor
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueResponse, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestBinarySensor:
    """Test class for BinarySensor objects."""

    #
    # TEST PROCESS
    #
    async def test_process(self) -> None:
        """Test process / reading telegrams from telegram queue."""
        xknx = XKNX()
        binaryinput = BinarySensor(xknx, "TestInput", "1/2/3")

        assert binaryinput.state is None

        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        binaryinput.process(telegram_on)
        assert binaryinput.state is True

        telegram_off = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        binaryinput.process(telegram_off)
        assert binaryinput.state is False

        binaryinput2 = BinarySensor(xknx, "TestInput", "1/2/4")
        assert binaryinput2.state is None

        telegram_off2 = Telegram(
            destination_address=GroupAddress("1/2/4"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        binaryinput2.process(telegram_off2)
        assert binaryinput2.last_telegram == telegram_off2
        assert binaryinput2.state is False

    async def test_process_invert(self) -> None:
        """Test process / reading telegrams from telegram queue."""
        xknx = XKNX()
        bs_invert = BinarySensor(xknx, "TestInput", "1/2/3", invert=True)

        assert bs_invert.state is None

        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        bs_invert.process(telegram_on)

        assert bs_invert.state is True

        telegram_off = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        bs_invert.process(telegram_off)
        assert bs_invert.state is False

    async def test_process_reset_after(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test process / reading telegrams from telegram queue."""
        xknx = XKNX()
        reset_after_sec = 1
        after_update_callback = Mock()
        binaryinput = BinarySensor(
            xknx,
            "TestInput",
            "1/2/3",
            reset_after=reset_after_sec,
            device_updated_cb=after_update_callback,
        )
        telegram_on = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        binaryinput.process(telegram_on)
        assert binaryinput.state

        await time_travel(reset_after_sec)
        assert not binaryinput.state
        # once for 'on' and once for 'off'
        assert after_update_callback.call_count == 2

        after_update_callback.reset_mock()
        # multiple telegrams during reset_after time period shall reset timer
        binaryinput.process(telegram_on)
        after_update_callback.assert_called_once()
        binaryinput.process(telegram_on)
        binaryinput.process(telegram_on)
        # second and third telegram resets timer but doesn't run callback
        after_update_callback.assert_called_once()
        assert binaryinput.state

        await time_travel(reset_after_sec)
        assert not binaryinput.state
        # once for 'on' and once for 'off'
        assert after_update_callback.call_count == 2

    async def test_process_wrong_payload(self) -> None:
        """Test process wrong telegram (wrong payload type)."""
        xknx = XKNX()
        binary_sensor = BinarySensor(xknx, "Warning", group_address_state="1/2/3")
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x1, 0x2, 0x3))),
        )
        with patch("logging.Logger.warning") as log_mock:
            binary_sensor.process(telegram)
            log_mock.assert_called_once()
            assert binary_sensor.state is None

    #
    # TEST SWITCHED ON
    #
    def test_is_on(self) -> None:
        """Test is_on() and is_off() of a BinarySensor with state 'on'."""
        xknx = XKNX()
        binaryinput = BinarySensor(xknx, "TestInput", "1/2/3")
        assert not binaryinput.is_on()
        assert binaryinput.is_off()
        binaryinput._set_internal_state(True)

        assert binaryinput.is_on()
        assert not binaryinput.is_off()

    #
    # TEST SWITCHED OFF
    #
    def test_is_off(self) -> None:
        """Test is_on() and is_off() of a BinarySensor with state 'off'."""
        xknx = XKNX()
        binaryinput = BinarySensor(xknx, "TestInput", "1/2/3")
        binaryinput._set_internal_state(False)

        assert not binaryinput.is_on()
        assert binaryinput.is_off()

    #
    # TEST PROCESS CALLBACK
    #
    async def test_process_callback(self) -> None:
        """Test after_update_callback after state of binary sensor was changed."""
        xknx = XKNX()
        switch = BinarySensor(
            xknx, "TestInput", group_address_state="1/2/3", ignore_internal_state=False
        )
        after_update_callback = Mock()

        switch.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        switch.process(telegram)
        # no _context_task started because ignore_internal_state is False
        assert switch._context_task is None
        after_update_callback.assert_called_once_with(switch)

        after_update_callback.reset_mock()
        # send same telegram again
        switch.process(telegram)
        after_update_callback.assert_not_called()

    async def test_process_callback_context_timeout(
        self, time_travel: EventLoopClockAdvancer
    ) -> None:
        """Test after_update_callback after context_timeout."""
        _timeout = 3
        xknx = XKNX()
        switch = BinarySensor(
            xknx,
            "TestInput",
            group_address_state="1/2/3",
            ignore_internal_state=True,
            context_timeout=_timeout,
        )
        after_update_callback = Mock()

        switch.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert switch.counter == 0

        switch.process(telegram)
        after_update_callback.assert_not_called()
        assert switch.counter == 1
        await time_travel(_timeout)
        after_update_callback.assert_called_with(switch)
        # once with counter 1 and once with counter 0
        assert after_update_callback.call_count == 2

        after_update_callback.reset_mock()
        # send same telegram again
        switch.process(telegram)
        assert switch.counter == 1
        await time_travel(_timeout / 2)  # not yet timed out
        switch.process(telegram)
        assert switch.counter == 2
        # incoming telegram resets timer (not sure if this is what we actually want)
        await time_travel(_timeout / 2)  # not yet timed out
        after_update_callback.assert_not_called()

        await time_travel(_timeout / 2)
        after_update_callback.assert_called_with(switch)
        # once with counter 2 and once with counter 0
        assert after_update_callback.call_count == 2
        assert switch.counter == 0

    async def test_process_callback_ignore_internal_state_no_counter(self) -> None:
        """Test after_update_callback after state of switch was changed."""
        xknx = XKNX()
        switch = BinarySensor(
            xknx,
            "TestInput",
            group_address_state="1/2/3",
            ignore_internal_state=True,
            context_timeout=0,
        )
        after_update_callback = Mock()

        switch.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        switch.process(telegram)
        # no _context_task started because context_timeout is False
        assert switch._context_task is None
        after_update_callback.assert_called_once_with(switch)

        after_update_callback.reset_mock()
        # send same telegram again
        switch.process(telegram)
        after_update_callback.assert_called_once_with(switch)

    async def test_process_group_value_response(self) -> None:
        """Test process of GroupValueResponse telegrams."""
        xknx = XKNX()
        switch = BinarySensor(
            xknx,
            "TestInput",
            group_address_state="1/2/3",
            ignore_internal_state=True,
        )
        after_update_callback = Mock()

        switch.register_device_updated_cb(after_update_callback)

        write_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        response_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(
                DPTBinary(1),
            ),
        )
        assert switch.state is None
        # initial GroupValueResponse changes state and runs callback
        switch.process(response_telegram)
        assert switch.state
        after_update_callback.assert_called_once_with(switch)
        # GroupValueWrite with same payload runs callback because of `ignore_internal_state`
        after_update_callback.reset_mock()
        switch.process(write_telegram)
        assert switch.state
        after_update_callback.assert_called_once_with(switch)
        # GroupValueResponse should not run callback when state has not changed
        after_update_callback.reset_mock()
        switch.process(response_telegram)
        after_update_callback.assert_not_called()

    async def test_process_always_callback(self) -> None:
        """Test process of GroupValueResponse telegrams."""
        xknx = XKNX()
        switch = BinarySensor(
            xknx,
            "TestInput",
            group_address_state="1/2/3",
            always_callback=True,
        )
        after_update_callback = Mock()

        switch.register_device_updated_cb(after_update_callback)

        write_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        response_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(
                DPTBinary(1),
            ),
        )
        assert switch.state is None
        # initial GroupValueResponse changes state and runs callback
        switch.process(response_telegram)
        assert switch.state
        after_update_callback.assert_called_once_with(switch)
        # GroupValueWrite with same payload runs callback because of `always_callback`
        after_update_callback.reset_mock()
        switch.process(write_telegram)
        assert switch.state
        after_update_callback.assert_called_once_with(switch)
        # GroupValueResponse should run callback because of `always_callback`
        after_update_callback.reset_mock()
        switch.process(response_telegram)
        after_update_callback.assert_called_once_with(switch)

    #
    # TEST COUNTER
    #
    def test_counter(self) -> None:
        """Test counter functionality."""
        xknx = XKNX()
        switch = BinarySensor(
            xknx, "TestInput", group_address_state="1/2/3", context_timeout=1
        )
        with patch("time.time") as mock_time:
            mock_time.return_value = 1517000000.0
            assert switch.bump_and_get_counter(True) == 1

            mock_time.return_value = 1517000000.1
            assert switch.bump_and_get_counter(True) == 2

            mock_time.return_value = 1517000000.2
            assert switch.bump_and_get_counter(False) == 1

            mock_time.return_value = 1517000000.3
            assert switch.bump_and_get_counter(True) == 3

            mock_time.return_value = 1517000000.4
            assert switch.bump_and_get_counter(False) == 2

            mock_time.return_value = 1517000002.0  # TIME OUT ...
            assert switch.bump_and_get_counter(True) == 1

            mock_time.return_value = 1517000004.1  # TIME OUT ...
            assert switch.bump_and_get_counter(False) == 1

    async def test_remove_tasks(self, xknx_no_interface: XKNX) -> None:
        """Test remove tasks."""
        xknx = xknx_no_interface
        switch = BinarySensor(
            xknx,
            "TestInput",
            group_address_state="1/2/3",
            context_timeout=1,
            reset_after=10,
        )
        xknx.devices.async_add(switch)
        async with xknx:
            write_telegram = Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(DPTBinary(1)),
            )
            switch.process(write_telegram)
            assert switch._context_task
            assert switch._reset_task
            xknx.devices.async_remove(switch)
            assert switch._context_task.done()
            assert switch._reset_task.done()



"""Unit test for KNX 4 byte float objects."""

import math
import struct
from unittest.mock import patch

import pytest

from xknx.dpt import (
    DPT4ByteFloat,
    DPTArray,
    DPTElectricCurrent,
    DPTElectricPotential,
    DPTFrequency,
    DPTPhaseAngleDeg,
    DPTPower,
)
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPT4ByteFloat:
    """Test class for KNX 4 byte/octet float object."""

    def test_4byte_float_values_from_power_meter(self) -> None:
        """Test parsing DPT4ByteFloat value from power meter."""
        assert DPT4ByteFloat.from_knx(DPTArray((0x43, 0xC6, 0x80, 00))) == 397
        assert DPT4ByteFloat.to_knx(397) == DPTArray((0x43, 0xC6, 0x80, 00))
        assert DPT4ByteFloat.from_knx(DPTArray((0x42, 0x38, 0x00, 00))) == 46
        assert DPT4ByteFloat.to_knx(46) == DPTArray((0x42, 0x38, 0x00, 00))

    def test_14_033(self) -> None:
        """Test parsing DPTFrequency unit."""
        assert DPTFrequency.unit == "Hz"

    def test_14_055(self) -> None:
        """Test DPTPhaseAngleDeg object."""
        assert DPT4ByteFloat.from_knx(DPTArray((0x42, 0xEF, 0x00, 0x00))) == 119.5
        assert DPT4ByteFloat.to_knx(119.5) == DPTArray((0x42, 0xEF, 0x00, 0x00))
        assert DPTPhaseAngleDeg.unit == "°"

    def test_14_057(self) -> None:
        """Test DPT4ByteFloat object."""
        assert DPT4ByteFloat.from_knx(DPTArray((0x3F, 0x71, 0xEB, 0x86))) == 0.9450001
        assert DPT4ByteFloat.to_knx(0.945000052452) == DPTArray(
            (0x3F, 0x71, 0xEB, 0x86)
        )
        assert DPT4ByteFloat.unit is None

    def test_4byte_float_values_from_voltage_meter(self) -> None:
        """Test parsing DPT4ByteFloat from voltage meter."""
        assert DPT4ByteFloat.from_knx(DPTArray((0x43, 0x65, 0xE3, 0xD7))) == 229.89
        assert DPT4ByteFloat.to_knx(229.89) == DPTArray((0x43, 0x65, 0xE3, 0xD7))

    def test_4byte_float_zero_value(self) -> None:
        """Test parsing and streaming of DPT4ByteFloat zero value."""
        assert DPT4ByteFloat.from_knx(DPTArray((0x00, 0x00, 0x00, 0x00))) == 0.00
        assert DPT4ByteFloat.to_knx(0.00) == DPTArray((0x00, 0x00, 0x00, 0x00))

    def test_4byte_float_special_value(self) -> None:
        """Test parsing and streaming of DPT4ByteFloat special value."""
        assert math.isnan(DPT4ByteFloat.from_knx(DPTArray((0x7F, 0xC0, 0x00, 0x00))))
        assert DPT4ByteFloat.to_knx(float("nan")) == DPTArray((0x7F, 0xC0, 0x00, 0x00))

        assert math.isinf(DPT4ByteFloat.from_knx(DPTArray((0x7F, 0x80, 0x00, 0x00))))
        assert DPT4ByteFloat.to_knx(float("inf")) == DPTArray((0x7F, 0x80, 0x00, 0x00))

        assert DPT4ByteFloat.from_knx(DPTArray((0xFF, 0x80, 0x00, 0x00))) == float(
            "-inf"
        )
        assert DPT4ByteFloat.to_knx(float("-inf")) == DPTArray((0xFF, 0x80, 0x00, 0x00))

        assert DPT4ByteFloat.from_knx(DPTArray((0x80, 0x00, 0x00, 0x00))) == float("-0")
        assert DPT4ByteFloat.to_knx(float("-0")) == DPTArray((0x80, 0x00, 0x00, 0x00))

    def test_4byte_float_to_knx_wrong_parameter(self) -> None:
        """Test parsing of DPT4ByteFloat with wrong value (string)."""
        with pytest.raises(ConversionError):
            DPT4ByteFloat.to_knx("fnord")

    def test_4byte_float_from_knx_wrong_parameter(self) -> None:
        """Test parsing of DPT4ByteFloat with wrong value (wrong number of bytes)."""
        with pytest.raises(CouldNotParseTelegram):
            DPT4ByteFloat.from_knx(DPTArray((0xF8, 0x01, 0x23)))

    def test_4byte_flaot_from_knx_unpack_error(self) -> None:
        """Test DPT4ByteFloat parsing with unpack error."""
        with patch("struct.unpack") as unpack_mock:
            unpack_mock.side_effect = struct.error()
            with pytest.raises(ConversionError):
                DPT4ByteFloat.from_knx(DPTArray((0x01, 0x23, 0x02, 0x02)))

    #
    # DPTElectricCurrent
    #
    def test_electric_current_settings(self) -> None:
        """Test attributes of DPTElectricCurrent."""
        assert DPTElectricCurrent.unit == "A"

    #
    # DPTElectricPotential
    #
    def test_electric_potential_settings(self) -> None:
        """Test attributes of DPTElectricPotential."""
        assert DPTElectricPotential.unit == "V"

    #
    # DPTPower
    #
    def test_power_settings(self) -> None:
        """Test attributes of DPTPower."""
        assert DPTPower.unit == "W"



"""Types used by XKNX."""

from collections.abc import Callable
import sys
from typing import TYPE_CHECKING, TypedDict, TypeVar

if sys.version_info >= (3, 11):
    from typing import Self as Self
else:
    from typing_extensions import Self as Self

if TYPE_CHECKING:
    from xknx.core.connection_manager import XknxConnectionState
    from xknx.devices import Device
    from xknx.telegram import Telegram

CallbackType = Callable[[], None]

ConnectionChangeCallbackType = Callable[["XknxConnectionState"], None]

DeviceT = TypeVar("DeviceT", bound="Device")
DeviceCallbackType = Callable[[DeviceT], None]

TelegramCallbackType = Callable[["Telegram"], None]


class DPTMainSubDict(TypedDict):
    """DPT type dictionary in accordance to xknxproject DPTType data."""

    main: int
    sub: int | None


DPTParsable = str | int | DPTMainSubDict



"""
Module for managing an DPT Step remote value.

DPT 1.007.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class RemoteValueStep(RemoteValue["RemoteValueStep.Direction"]):
    """Abstraction for remote value of KNX DPT 1.007 / DPT_Step."""

    class Direction(Enum):
        """Enum for indicating the direction."""

        DECREASE = 0
        INCREASE = 1

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        device_name: str | None = None,
        feature_name: str = "Step",
        after_update_cb: RVCallbackType[Direction] | None = None,
        invert: bool = False,
    ) -> None:
        """Initialize remote value of KNX DPT 1.007."""
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )
        self.invert = invert

    # from KNX Association System Specifications AS v1.5.00:
    # 1.007 DPT_Step   0 = Decrease 1 = Increase
    # 1.008 DPT_UpDown 0 = Up       1 = Down

    def to_knx(self, value: RemoteValueStep.Direction) -> DPTBinary:
        """Convert value to payload."""
        if value == self.Direction.INCREASE:
            return DPTBinary(0) if self.invert else DPTBinary(1)
        if value == self.Direction.DECREASE:
            return DPTBinary(1) if self.invert else DPTBinary(0)
        raise ConversionError(
            "value invalid",
            value=value,
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def from_knx(self, payload: DPTArray | DPTBinary) -> RemoteValueStep.Direction:
        """Convert current payload to value."""
        if payload.value == 1:
            return self.Direction.DECREASE if self.invert else self.Direction.INCREASE
        if payload.value == 0:
            return self.Direction.INCREASE if self.invert else self.Direction.DECREASE
        raise CouldNotParseTelegram(
            "Payload invalid",
            payload=str(payload),
            device_name=self.device_name,
            feature_name=self.feature_name,
        )

    def increase(self) -> None:
        """Increase value."""
        self.set(self.Direction.INCREASE)

    def decrease(self) -> None:
        """Decrease the value."""
        self.set(self.Direction.DECREASE)



"""
Module for handling KNX primitives.

* KNX Addresses
* KNX Telegrams

"""

from .address import GroupAddress, GroupAddressType, IndividualAddress
from .address_filter import AddressFilter
from .telegram import Telegram, TelegramDecodedData, TelegramDirection

__all__ = [
    "AddressFilter",
    "GroupAddress",
    "GroupAddressType",
    "IndividualAddress",
    "Telegram",
    "TelegramDecodedData",
    "TelegramDirection",
]



"""Example for GatewayScanner."""

import asyncio

from xknx import XKNX
from xknx.io import GatewayScanner


async def main() -> None:
    """Search for available KNX/IP devices with GatewayScanner and print out result if a device was found."""
    xknx = XKNX()
    gatewayscanner = GatewayScanner(xknx)

    async for gateway in gatewayscanner.async_scan():
        print(f"{gateway.individual_address} {gateway.name}")
        print(f"  {gateway.ip_addr}:{gateway.port}")
        tunnelling = (
            "Secure"
            if gateway.tunnelling_requires_secure
            else "TCP"
            if gateway.supports_tunnelling_tcp
            else "UDP"
            if gateway.supports_tunnelling
            else "No"
        )
        print(f"  Tunnelling: {tunnelling}")
        routing = (
            "Secure"
            if gateway.routing_requires_secure
            else "Yes"
            if gateway.supports_routing
            else "No"
        )
        print(f"  Routing: {routing}")

    if not gatewayscanner.found_gateways:
        print("No Gateways found")


asyncio.run(main())



"""Test xknx tools package."""

from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from xknx import XKNX
from xknx.devices import NumericValue
from xknx.dpt import DPTArray, DPTBinary, DPTTemperature
from xknx.exceptions import ConversionError
from xknx.telegram import GroupAddress, Telegram, TelegramDirection, apci
from xknx.tools import (
    group_value_read,
    group_value_response,
    group_value_write,
    read_group_value,
)


def test_group_value_read() -> None:
    """Test group_value_read."""
    xknx = XKNX()
    group_address = "1/2/3"

    read = Telegram(
        destination_address=GroupAddress(group_address),
        payload=apci.GroupValueRead(),
    )
    group_value_read(xknx, "1/2/3")
    assert xknx.telegrams.qsize() == 1
    assert xknx.telegrams.get_nowait() == read


def test_group_value_response() -> None:
    """Test group_value_response."""
    xknx = XKNX()
    group_address = "1/2/3"

    response = Telegram(
        destination_address=GroupAddress(group_address),
        payload=apci.GroupValueResponse(DPTBinary(1)),
    )
    group_value_response(xknx, "1/2/3", True)
    assert xknx.telegrams.qsize() == 1
    assert xknx.telegrams.get_nowait() == response


@pytest.mark.parametrize(
    ("value", "value_type", "expected"),
    [
        (50, "percent", DPTArray((0x80,))),
        (True, None, DPTBinary(1)),
        (False, None, DPTBinary(0)),
        (20.48, DPTTemperature, DPTArray((0x0C, 0x00))),
        (-100, 6, DPTArray((0x9C,))),
        ((0x0C, 0x00), None, DPTArray((0x0C, 0x00))),
        (DPTBinary(1), None, DPTBinary(1)),
    ],
)
def test_group_value_write(
    value: Any, value_type: Any, expected: DPTArray | DPTBinary
) -> None:
    """Test group_value_write."""
    xknx = XKNX()
    group_address = "1/2/3"

    write = Telegram(
        destination_address=GroupAddress(group_address),
        payload=apci.GroupValueWrite(expected),
    )
    group_value_write(xknx, "1/2/3", value, value_type=value_type)
    assert xknx.telegrams.qsize() == 1
    assert xknx.telegrams.get_nowait() == write


@pytest.mark.parametrize(
    ("value", "value_type", "error_type"),
    [
        (50, "unknown", ValueError),
        (50, 9.001, ValueError),  # float is invalid
        (101, "percent", ConversionError),  # too big
    ],
)
def test_group_value_write_invalid(
    value: int, value_type: Any, error_type: type[Exception]
) -> None:
    """Test group_value_write."""
    xknx = XKNX()
    with pytest.raises(error_type):
        group_value_write(xknx, "1/2/3", value, value_type=value_type)


@pytest.mark.parametrize(
    ("value", "value_type", "expected"),
    [
        (50, "percent", DPTArray((0x80,))),
        ((0x80,), None, DPTArray((0x80,))),
        (True, None, DPTBinary(1)),
    ],
)
@patch("xknx.core.value_reader.ValueReader.read")
async def test_read_group_value(
    value_reader_read_mock: MagicMock,
    value: Any,
    value_type: Any,
    expected: DPTArray | DPTBinary,
) -> None:
    """Test read_group_value."""
    xknx = XKNX()
    test_group_address = "1/2/3"

    response_telegram = Telegram(
        destination_address=GroupAddress(test_group_address),
        direction=TelegramDirection.INCOMING,
        payload=apci.GroupValueResponse(expected),
    )
    value_reader_read_mock.return_value = response_telegram

    response_value = await read_group_value(
        xknx, test_group_address, value_type=value_type
    )
    # GroupValueRead telegram is not in queue because ValueReader.read is mocked.
    # This is tested in ValueReader tests.
    assert response_value == value


async def test_tools_with_internal_addresses(xknx_no_interface: XKNX) -> None:
    """Test tools using internal addresses."""
    xknx = xknx_no_interface
    await xknx.start()

    internal_address = "i-test"
    test_type = "1byte_unsigned"
    number = NumericValue(
        xknx,
        "Test",
        group_address=internal_address,
        value_type=test_type,
        respond_to_read=True,
    )
    xknx.devices.async_add(number)

    assert number.resolve_state() is None
    group_value_write(xknx, internal_address, 1, value_type=test_type)
    await xknx.telegrams.join()
    assert number.resolve_state() == 1

    response_value = await read_group_value(
        xknx, internal_address, value_type=test_type
    )
    assert response_value == 1
    await xknx.stop()



"""Implementation of Basic KNX Time."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
import datetime
from typing import Any

from xknx.exceptions import ConversionError

from .dpt import DPTComplex, DPTComplexData, DPTEnumData
from .payload import DPTArray, DPTBinary


class KNXDay(DPTEnumData):
    """Enum for the different KNX days."""

    NO_DAY = 0
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7


@dataclass(slots=True)
class KNXTime(DPTComplexData):
    """Class for KNX Time."""

    hour: int
    minutes: int
    seconds: int
    day: KNXDay = KNXDay.NO_DAY

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> KNXTime:
        """Init from a dictionary."""
        try:
            hour = int(data["hour"])
            minutes = int(data["minutes"])
            seconds = int(data["seconds"])
            day = KNXDay.parse(data.get("day", KNXDay.NO_DAY))
        except (KeyError, TypeError, ValueError) as err:
            raise ValueError(f"Invalid value for KNXTime: {err}") from err
        return cls(hour=hour, minutes=minutes, seconds=seconds, day=day)

    def as_dict(self) -> dict[str, int | str]:
        """Create a JSON serializable dictionary."""
        return {
            "hour": self.hour,
            "minutes": self.minutes,
            "seconds": self.seconds,
            "day": self.day.name.lower(),
        }

    def as_time(self) -> datetime.time:
        """Return time object. Ignoring day field."""
        return datetime.time(self.hour, self.minutes, self.seconds)

    @classmethod
    def from_time(cls, time: datetime.time) -> KNXTime:
        """Return KNXTime object from time object. Day field is set to NO_DAY."""
        return cls(time.hour, time.minute, time.second)


class DPTTime(DPTComplex[KNXTime]):
    """
    Abstraction for KNX 3 Octet Time.

    DPT 10.001
    """

    data_type = KNXTime
    payload_type = DPTArray
    payload_length = 3
    dpt_main_number = 10
    dpt_sub_number = 1
    value_type = "time"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> KNXTime:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)

        weekday = (raw[0] & 0xE0) >> 5  # can not be out of range - 3 bits 0..7
        hours = raw[0] & 0x1F
        minutes = raw[1] & 0x3F
        seconds = raw[2] & 0x3F
        try:
            DPTTime._test_range(hours, minutes, seconds)
        except ValueError as err:
            raise ConversionError(f"Could not parse {cls.dpt_name()}: {err}") from err
        return KNXTime(
            hour=hours, minutes=minutes, seconds=seconds, day=KNXDay(weekday)
        )

    @classmethod
    def _to_knx(cls, value: KNXTime) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            DPTTime._test_range(value.hour, value.minutes, value.seconds)
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}: {err}"
            ) from err
        return DPTArray(
            (
                value.day.value << 5 | value.hour,
                value.minutes,
                value.seconds,
            )
        )

    @staticmethod
    def _test_range(hour: int, minutes: int, seconds: int) -> None:
        """Test if values are in the correct value range."""
        if not 0 <= hour <= 23:
            raise ValueError(f"Hour out of range 0..23: {hour}")
        if not 0 <= minutes <= 59:
            raise ValueError(f"Minutes out of range 0..59: {minutes}")
        if not 0 <= seconds <= 59:
            raise ValueError(f"Seconds out of range 0..59: {seconds}")



"""Unit test for keyring reader."""

from pathlib import Path

import pytest

from xknx.exceptions.exception import InvalidSecureConfiguration
from xknx.secure.keyring import (
    InterfaceType,
    Keyring,
    XMLDevice,
    XMLInterface,
    sync_load_keyring,
    verify_keyring_signature,
)
from xknx.telegram import GroupAddress, IndividualAddress


class TestKeyRing:
    """Test class for keyring."""

    keyring_test_file = Path(__file__).parent / "resources/keyring.knxkeys"
    testcase_file: str = Path(__file__).parent / "resources/testcase.knxkeys"
    special_chars_file = (
        Path(__file__).parent / "resources/special_chars_secure_tunnel.knxkeys"
    )
    data_secure_ip = (
        Path(__file__).parent / "resources/DataSecure_only_one_interface.knxkeys"
    )
    data_secure_usb = Path(__file__).parent / "resources/DataSecure_usb.knxkeys"

    @staticmethod
    def assert_interface(
        keyring: Keyring, password: str, ia: IndividualAddress
    ) -> None:
        """Verify password for given user."""
        matched = False
        if interface := keyring.get_tunnel_interface_by_individual_address(ia):
            matched = True
            assert interface.decrypted_password == password

        assert matched

    def test_load_keyring_test(self) -> None:
        """Test load keyring from knxkeys file."""
        keyring = sync_load_keyring(self.keyring_test_file, "pwd")
        TestKeyRing.assert_interface(keyring, "user4", IndividualAddress("1.1.4"))
        TestKeyRing.assert_interface(keyring, "@zvI1G&_", IndividualAddress("1.1.6"))
        TestKeyRing.assert_interface(keyring, "ZvDY-:g#", IndividualAddress("1.1.7"))
        TestKeyRing.assert_interface(keyring, "user2", IndividualAddress("1.1.2"))
        assert keyring.backbone.multicast_address == "224.0.23.12"
        assert keyring.backbone.latency == 1000
        assert keyring.backbone.decrypted_key == bytes.fromhex(
            "96f034fccf510760cbd63da0f70d4a9d"
        )

    def test_load_testcase_file(self) -> None:
        """Test load keyring from knxkeys file."""
        keyring = sync_load_keyring(self.testcase_file, "password")
        TestKeyRing.assert_interface(keyring, "user1", IndividualAddress("1.0.1"))
        TestKeyRing.assert_interface(keyring, "user2", IndividualAddress("1.0.11"))
        TestKeyRing.assert_interface(keyring, "user3", IndividualAddress("1.0.12"))
        TestKeyRing.assert_interface(keyring, "user4", IndividualAddress("1.0.13"))
        assert keyring.devices[0].decrypted_management_password == "commissioning"
        assert keyring.backbone.multicast_address == "224.0.23.12"
        assert keyring.backbone.latency == 1000
        assert keyring.backbone.decrypted_key == bytes.fromhex(
            "cf89fd0f18f4889783c7ef44ee1f5e14"
        )

        interface: XMLInterface = keyring.interfaces[0]
        device: XMLDevice = keyring.get_device_by_interface(interface)
        assert device is not None
        assert device.decrypted_authentication == "authenticationcode"

    def test_load_special_chars_file(self) -> None:
        """Test load keyring from knxkeys file."""
        keyring = sync_load_keyring(self.special_chars_file, "test")
        TestKeyRing.assert_interface(keyring, "tunnel_2", IndividualAddress("1.0.2"))
        TestKeyRing.assert_interface(keyring, "tunnel_3", IndividualAddress("1.0.3"))
        TestKeyRing.assert_interface(keyring, "tunnel_4", IndividualAddress("1.0.4"))
        TestKeyRing.assert_interface(keyring, "tunnel_5", IndividualAddress("1.0.5"))
        TestKeyRing.assert_interface(keyring, "tunnel_6", IndividualAddress("1.0.6"))
        assert keyring.backbone is None

    def test_load_data_secure_ip(self) -> None:
        """Test load keyring from knxkeys file."""
        keyring = sync_load_keyring(self.data_secure_ip, "test")
        assert len(keyring.interfaces) == 1
        tunnel = keyring.interfaces[0]
        assert tunnel is not None
        assert tunnel.password is None
        assert tunnel.decrypted_password is None
        assert tunnel.user_id is None
        assert tunnel.type is InterfaceType.TUNNELING
        assert len(tunnel.group_addresses) == 3
        assert tunnel.group_addresses[GroupAddress("0/0/1")] == [
            IndividualAddress("1.0.1"),
            IndividualAddress("1.0.2"),
        ]
        assert tunnel.group_addresses[GroupAddress("0/0/3")] == [
            IndividualAddress("1.0.1"),
            IndividualAddress("1.0.2"),
        ]
        assert tunnel.group_addresses[GroupAddress("31/7/255")] == []
        assert keyring.backbone is None

    def test_load_data_secure_usb(self) -> None:
        """Test load keyring from knxkeys file."""
        keyring = sync_load_keyring(self.data_secure_usb, "test")
        assert len(keyring.interfaces) == 1
        interface = keyring.interfaces[0]
        assert interface is not None
        assert interface.password is None
        assert interface.decrypted_password is None
        assert interface.user_id is None
        assert interface.host is None
        assert interface.type is InterfaceType.USB
        assert len(interface.group_addresses) == 1
        assert interface.group_addresses[GroupAddress("31/7/255")] == [
            IndividualAddress("1.0.4")
        ]
        assert keyring.backbone is None

    def test_verify_signature(self) -> None:
        """Test signature verification."""
        assert verify_keyring_signature(self.keyring_test_file, "pwd")
        assert verify_keyring_signature(self.testcase_file, "password")
        assert verify_keyring_signature(self.special_chars_file, "test")

    def test_invalid_signature(self) -> None:
        """Test invalid signature throws error."""
        with pytest.raises(InvalidSecureConfiguration):
            sync_load_keyring(self.testcase_file, "wrong_password")

    def test_raises_error(self) -> None:
        """Test raises error if password is wrong."""
        with pytest.raises(InvalidSecureConfiguration):
            sync_load_keyring(
                self.testcase_file, "wrong_password", validate_signature=False
            )

    def test_keyring_get_methods_full(self) -> None:
        """Test keyring get_* methods for full project export."""
        keyring = sync_load_keyring(self.keyring_test_file, "pwd")
        test_interfaces = keyring.get_tunnel_interfaces_by_host(
            host=IndividualAddress("1.1.10")
        )
        assert len(test_interfaces) == 1
        test_interface = test_interfaces[0]

        test_device = keyring.get_device_by_interface(interface=test_interface)
        assert test_device.individual_address == IndividualAddress("1.1.10")

        test_host = keyring.get_tunnel_host_by_interface(
            tunnelling_slot=IndividualAddress("1.1.8")
        )
        assert test_host == IndividualAddress("1.1.0")

        test_host = keyring.get_tunnel_host_by_interface(
            tunnelling_slot=IndividualAddress("1.1.10")
        )
        assert test_host is None

        test_interface = keyring.get_tunnel_interface_by_host_and_user_id(
            host=IndividualAddress("1.1.0"), user_id=4
        )
        assert test_interface.individual_address == IndividualAddress("1.1.7")

        test_interface = keyring.get_tunnel_interface_by_individual_address(
            tunnelling_slot=IndividualAddress("1.1.8")
        )
        assert test_interface.user_id == 8

        test_interface = keyring.get_tunnel_interface_by_individual_address(
            tunnelling_slot=IndividualAddress("1.1.20")
        )
        assert test_interface.user_id is None
        assert test_interface.host == IndividualAddress("1.1.10")
        # this doesn't check for `type`, but there are no other than TUNNELLING interfaces in this keyring
        test_interface = keyring.get_interface_by_individual_address(
            individual_address=IndividualAddress("1.1.20")
        )
        assert test_interface.host == IndividualAddress("1.1.10")

        full_ga_key_table = keyring.get_data_secure_group_keys()
        assert len(full_ga_key_table) == 1

        individual_ga_key_table = keyring.get_data_secure_group_keys(
            receiver=IndividualAddress("1.1.7")
        )
        assert len(individual_ga_key_table) == 0

        ia_seq_nums = keyring.get_data_secure_senders()
        assert len(ia_seq_nums) == 5

    def test_keyring_get_methods_one_interface(self) -> None:
        """Test keyring get_* methods for partial export."""
        keyring = sync_load_keyring(self.data_secure_ip, "test")

        full_ga_key_table = keyring.get_data_secure_group_keys()
        assert len(full_ga_key_table) == 3

        individual_ga_key_table = keyring.get_data_secure_group_keys(
            receiver=IndividualAddress("1.0.4")
        )
        assert len(individual_ga_key_table) == 3

        ia_seq_nums = keyring.get_data_secure_senders()
        assert ia_seq_nums == {
            IndividualAddress("1.0.1"): 0,
            IndividualAddress("1.0.2"): 0,
        }

    def test_keyring_metadata(self) -> None:
        """Test keyring metadata parsing."""
        keyring = sync_load_keyring(self.data_secure_ip, "test")
        assert keyring.project_name == "DataSecure_only"
        assert keyring.created_by == "ETS 5.7.7 (Build 1428)"
        assert keyring.created == "2023-02-06T21:17:09"
        assert keyring.xmlns == "http://knx.org/xml/keyring/1"



"""Unit test for KNX/IP HPAI objects."""

import pytest

from xknx.exceptions import ConversionError, CouldNotParseKNXIP
from xknx.knxip import HPAI, HostProtocol


class TestKNXIPHPAI:
    """Test class for KNX/IP HPAI objects."""

    def test_hpai_udp(self) -> None:
        """Test parsing and streaming HPAI KNX/IP fragment."""
        raw = bytes((0x08, 0x01, 0xC0, 0xA8, 0x2A, 0x01, 0x84, 0x95))

        hpai = HPAI()
        assert hpai.from_knx(raw) == 8
        assert hpai.ip_addr == "192.168.42.1"
        assert hpai.port == 33941
        assert hpai.protocol == HostProtocol.IPV4_UDP
        assert not hpai.route_back

        hpai2 = HPAI(ip_addr="192.168.42.1", port=33941)
        assert hpai2.to_knx() == raw
        assert not hpai2.route_back

    def test_hpai_tcp(self) -> None:
        """Test parsing and streaming HPAI KNX/IP fragment."""
        raw = bytes((0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))

        hpai = HPAI()
        assert hpai.from_knx(raw) == 8
        assert hpai.ip_addr == "0.0.0.0"
        assert hpai.port == 0
        assert hpai.protocol == HostProtocol.IPV4_TCP
        assert hpai.route_back

        hpai2 = HPAI(ip_addr="0.0.0.0", port=0, protocol=HostProtocol.IPV4_TCP)
        assert hpai2.to_knx() == raw

    def test_from_knx_wrong_input1(self) -> None:
        """Test parsing of wrong HPAI KNX/IP packet (wrong length)."""
        raw = bytes((0x08, 0x01, 0xC0, 0xA8, 0x2A))
        with pytest.raises(CouldNotParseKNXIP):
            HPAI().from_knx(raw)

    def test_from_knx_wrong_input2(self) -> None:
        """Test parsing of wrong HPAI KNX/IP packet (wrong length byte)."""
        raw = bytes((0x09, 0x01, 0xC0, 0xA8, 0x2A, 0x01, 0x84, 0x95))
        with pytest.raises(CouldNotParseKNXIP):
            HPAI().from_knx(raw)

    def test_from_knx_wrong_input3(self) -> None:
        """Test parsing of wrong HPAI KNX/IP packet (wrong HPAI type)."""
        raw = bytes((0x08, 0x03, 0xC0, 0xA8, 0x2A, 0x01, 0x84, 0x95))
        with pytest.raises(CouldNotParseKNXIP):
            HPAI().from_knx(raw)

    def test_to_knx_wrong_ip(self) -> None:
        """Test serializing HPAI to KNV/IP with wrong ip type."""
        hpai = HPAI(ip_addr=127001)
        with pytest.raises(ConversionError):
            hpai.to_knx()

    def test_route_back(self) -> None:
        """Test route_back property."""
        hpai = HPAI()
        assert hpai.ip_addr == "0.0.0.0"
        assert hpai.port == 0
        assert hpai.route_back
        hpai.ip_addr = "10.1.2.3"
        assert not hpai.route_back



"""Abstraction to send a TunnelingRequest and wait for TunnelingResponse."""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.knxip import KNXIPFrame, TunnellingAck, TunnellingRequest

from .request_response import RequestResponse

if TYPE_CHECKING:
    from xknx.io.transport import UDPTransport


class Tunnelling(RequestResponse):
    """Class to send TunnelingRequest and wait for TunnelingACK (UDP only)."""

    transport: UDPTransport

    def __init__(
        self,
        transport: UDPTransport,
        data_endpoint: tuple[str, int] | None,
        tunnelling_request: TunnellingRequest,
    ) -> None:
        """Initialize Tunnelling class."""
        self.data_endpoint_addr = data_endpoint
        self.tunnelling_request = tunnelling_request
        super().__init__(transport, TunnellingAck)

    async def send_request(self) -> None:
        """Build knxipframe (within derived class) and send via UDP."""
        self.transport.send(self.create_knxipframe(), addr=self.data_endpoint_addr)

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        return KNXIPFrame.init_from_body(self.tunnelling_request)



"""Module for handling devices like Lights, Switches or Covers."""

from .binary_sensor import BinarySensor
from .climate import Climate
from .climate_mode import ClimateMode
from .cover import Cover
from .datetime import DateDevice, DateTimeDevice, TimeDevice
from .device import Device
from .devices import Devices
from .expose_sensor import ExposeSensor
from .fan import Fan
from .light import Light
from .notification import Notification
from .numeric_value import NumericValue
from .raw_value import RawValue
from .scene import Scene
from .sensor import Sensor
from .switch import Switch
from .travelcalculator import TravelCalculator, TravelStatus
from .weather import Weather

__all__ = [
    "BinarySensor",
    "Climate",
    "ClimateMode",
    "Cover",
    "DateDevice",
    "DateTimeDevice",
    "Device",
    "Devices",
    "ExposeSensor",
    "Fan",
    "Light",
    "Notification",
    "NumericValue",
    "RawValue",
    "Scene",
    "Sensor",
    "Switch",
    "TimeDevice",
    "TravelCalculator",
    "TravelStatus",
    "Weather",
]



"""Unit test for telegram received callback."""

import asyncio
from unittest.mock import AsyncMock, MagicMock, Mock, call, patch

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import CommunicationError, CouldNotParseTelegram
from xknx.telegram import AddressFilter, Telegram, TelegramDirection
from xknx.telegram.address import GroupAddress, IndividualAddress, InternalGroupAddress
from xknx.telegram.apci import GroupValueWrite


class TestTelegramQueue:
    """Test class for telegram queue."""

    #
    # TEST START, RUN, STOP
    #
    async def test_start(self) -> None:
        """Test start, run and stop."""

        xknx = XKNX()

        telegram_in = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        await xknx.telegram_queue.start()

        assert not xknx.telegram_queue._consumer_task.done()
        # queue shall now consume telegrams from xknx.telegrams
        assert xknx.telegrams.qsize() == 0
        xknx.telegrams.put_nowait(telegram_in)
        xknx.telegrams.put_nowait(telegram_in)
        assert xknx.telegrams.qsize() == 2
        # wait until telegrams are consumed
        await xknx.telegrams.join()
        assert xknx.telegrams.qsize() == 0
        await xknx.telegrams.join()
        assert xknx.telegrams.qsize() == 0
        # stop run() task with stop()
        await xknx.telegram_queue.stop()
        assert xknx.telegram_queue._consumer_task.done()

    @patch("asyncio.sleep", new_callable=AsyncMock)
    async def test_rate_limit(self, async_sleep_mock: AsyncMock) -> None:
        """Test rate limit."""
        xknx = XKNX(
            rate_limit=20,  # 50 ms per outgoing telegram
        )
        sleep_time = 0.05  # 1 / 20

        telegram_in = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        telegram_out = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram_internal = Telegram(
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
            destination_address=InternalGroupAddress("i-test"),
        )
        await xknx.telegram_queue.start()

        # no sleep for incoming telegrams
        xknx.telegrams.put_nowait(telegram_in)
        xknx.telegrams.put_nowait(telegram_in)
        await xknx.telegrams.join()
        assert async_sleep_mock.call_count == 0

        # sleep for outgoing telegrams
        xknx.telegrams.put_nowait(telegram_out)
        xknx.telegrams.put_nowait(telegram_out)
        await xknx.telegrams.join()
        assert async_sleep_mock.call_count == 2
        async_sleep_mock.assert_called_with(sleep_time)

        async_sleep_mock.reset_mock()
        # no sleep for internal group address telegrams
        xknx.telegrams.put_nowait(telegram_internal)
        xknx.telegrams.put_nowait(telegram_internal)
        await xknx.telegrams.join()
        async_sleep_mock.assert_not_called()

        await xknx.telegram_queue.stop()

    #
    # TEST REGISTER
    #
    async def test_register(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""

        xknx = XKNX()
        telegram_received_cb = Mock()
        xknx.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        telegram_received_cb.assert_called_once_with(telegram)

    async def test_register_with_outgoing_telegrams(self) -> None:
        """Test telegram_received_callback with outgoing telegrams."""

        xknx = XKNX()
        xknx.cemi_handler = AsyncMock()
        telegram_received_cb = Mock()
        xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb, None, None, True
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        await xknx.telegram_queue.process_telegram_outgoing(telegram)
        telegram_received_cb.assert_called_once_with(telegram)

    async def test_register_with_outgoing_telegrams_does_not_trigger(self) -> None:
        """Test telegram_received_callback with outgoing telegrams."""

        xknx = XKNX()
        xknx.cemi_handler = AsyncMock()
        telegram_received_cb = Mock()
        xknx.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        await xknx.telegram_queue.process_telegram_outgoing(telegram)
        telegram_received_cb.assert_not_called()

    #
    # TEST UNREGISTER
    #
    async def test_unregister(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""

        xknx = XKNX()
        telegram_received_cb = Mock()

        callback = xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb
        )
        xknx.telegram_queue.unregister_telegram_received_cb(callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        telegram_received_cb.assert_not_called()

    #
    # TEST PROCESS
    #
    @patch("xknx.devices.Devices.devices_by_group_address")
    async def test_process_to_device(self, devices_by_ga_mock: Mock) -> None:
        """Test process_telegram_incoming for forwarding telegram to a device."""

        xknx = XKNX()
        test_device = Mock()
        devices_by_ga_mock.return_value = [test_device]

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        devices_by_ga_mock.assert_called_once_with(GroupAddress("1/2/3"))
        test_device.process.assert_called_once_with(telegram)

    @patch("xknx.devices.Devices.process")
    async def test_process_to_callback(self, devices_process: MagicMock) -> None:
        """Test process_telegram_incoming with callback."""
        xknx = XKNX()
        telegram_received_cb = Mock()

        xknx.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        telegram_received_cb.assert_called_once_with(telegram)
        devices_process.assert_called_once_with(telegram)

    async def test_callback_decoded_telegram_data(self) -> None:
        """Test telegram_received_callback having decoded telegram data."""

        xknx = XKNX()
        xknx.group_address_dpt.set({"1/2/3": {"main": 5, "sub": 1}})
        telegram_received_cb = Mock()
        xknx.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(
                DPTArray(
                    0x7F,
                )
            ),
        )
        await xknx.telegram_queue.start()
        xknx.telegrams.put_nowait(telegram)
        await xknx.telegrams.join()
        await xknx.telegram_queue.stop()

        assert telegram_received_cb.call_count == 1
        received = telegram_received_cb.call_args_list[0][0][0]
        assert received == telegram
        assert received.decoded_data is not None
        assert received.decoded_data.value == 50

    async def test_outgoing(self) -> None:
        """Test outgoing telegrams in telegram queue."""
        xknx = XKNX()

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        # log a warning if there is no KNXIP interface instantiated
        with pytest.raises(CommunicationError):  # from xknx.knxip_interface.send_cemi
            await xknx.telegram_queue.process_telegram_outgoing(telegram)

    @patch("logging.Logger.error")
    @patch("xknx.core.TelegramQueue.process_telegram_incoming", new_callable=MagicMock)
    async def test_process_exception(
        self, process_tg_in_mock: MagicMock, logging_error_mock: MagicMock
    ) -> None:
        """Test process_telegram exception handling."""

        xknx = XKNX()

        async def process_exception() -> None:
            raise CouldNotParseTelegram(
                "Something went wrong when receiving the telegram."
            )

        process_tg_in_mock.return_value = asyncio.ensure_future(process_exception())

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        xknx.telegrams.put_nowait(telegram)
        await xknx.telegram_queue._process_all_telegrams()

        logging_error_mock.assert_called_once_with(
            "Error while processing telegram %s",
            CouldNotParseTelegram("Something went wrong when receiving the telegram."),
        )

    @patch("xknx.core.TelegramQueue.process_telegram_outgoing", new_callable=AsyncMock)
    @patch("xknx.core.TelegramQueue.process_telegram_incoming", new_callable=AsyncMock)
    async def test_process_all_telegrams(
        self,
        process_telegram_incoming_mock: AsyncMock,
        process_telegram_outgoing_mock: AsyncMock,
    ) -> None:
        """Test _process_all_telegrams for clearing the queue."""
        xknx = XKNX()

        telegram_in = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram_out = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(1)),
        )

        xknx.telegrams.put_nowait(telegram_in)
        xknx.telegrams.put_nowait(telegram_out)
        await xknx.telegram_queue._process_all_telegrams()

        process_telegram_incoming_mock.assert_called_once()
        process_telegram_outgoing_mock.assert_called_once()

    #
    # TEST NO FILTERS
    #
    async def test_callback_no_filters(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""
        xknx = XKNX()
        telegram_received_cb = Mock()

        xknx.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        xknx.telegrams.put_nowait(telegram)
        await xknx.telegram_queue._process_all_telegrams()

        telegram_received_cb.assert_called_with(telegram)

    #
    # TEST POSITIVE FILTERS
    #
    async def test_callback_positive_address_filters(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""
        xknx = XKNX()
        telegram_received_cb = Mock()

        xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb,
            address_filters=[AddressFilter("2/4-8/*"), AddressFilter("1/2/-8")],
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        xknx.telegrams.put_nowait(telegram)
        await xknx.telegram_queue._process_all_telegrams()

        telegram_received_cb.assert_called_with(telegram)

    #
    # TEST NEGATIVE FILTERS
    #
    async def test_callback_negative_address_filters(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""
        xknx = XKNX()
        telegram_received_cb = Mock()

        xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb,
            address_filters=[AddressFilter("2/4-8/*"), AddressFilter("1/2/8-")],
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        xknx.telegrams.put_nowait(telegram)
        await xknx.telegram_queue._process_all_telegrams()

        telegram_received_cb.assert_not_called()

    async def test_callback_group_addresses(self) -> None:
        """Test telegram_received_callback after state of switch was changed."""
        xknx = XKNX()
        telegram_received_cb_one = Mock()
        telegram_received_cb_two = Mock()

        callback_one = xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb_one,
            address_filters=[],
            group_addresses=[GroupAddress("1/2/3")],
        )
        callback_two = xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_cb_two, address_filters=[], group_addresses=[]
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        telegram_received_cb_one.assert_called_once_with(telegram)
        telegram_received_cb_two.assert_not_called()

        telegram_received_cb_one.reset_mock()
        # modify the filters - add/remove a GroupAddress
        callback_one.group_addresses.remove(GroupAddress("1/2/3"))
        callback_two.group_addresses.append(GroupAddress("1/2/3"))
        await xknx.telegram_queue.process_telegram_incoming(telegram)
        telegram_received_cb_one.assert_not_called()
        telegram_received_cb_two.assert_called_once_with(telegram)

    #
    # TEST EXCEPTION HANDLING
    #
    @patch("logging.Logger.exception")
    @patch("xknx.xknx.Devices.process", side_effect=Exception)
    async def test_process_raising(
        self, process_mock: MagicMock, logging_exception_mock: MagicMock
    ) -> None:
        """Test unexpected exception handling in telegram queues."""
        xknx = XKNX()
        telegram_in = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        # InternalGroupAddress to avoid CommunicationError for missing knxip_interface
        telegram_out = Telegram(
            destination_address=InternalGroupAddress("i-outgoing"),
            direction=TelegramDirection.OUTGOING,
            payload=GroupValueWrite(DPTBinary(0)),
        )
        xknx.telegrams.put_nowait(telegram_in)
        xknx.telegrams.put_nowait(telegram_out)

        await xknx.telegram_queue.start()
        await xknx.telegram_queue.stop()

        log_calls = [
            call(
                "Unexpected error while processing incoming telegram %s",
                telegram_in,
            ),
            call(
                "Unexpected error while processing outgoing telegram %s",
                telegram_out,
            ),
        ]
        logging_exception_mock.assert_has_calls(log_calls)

    @patch("logging.Logger.exception")
    async def test_callback_raising(self, logging_exception_mock: MagicMock) -> None:
        """Test telegram_received_callback raising an exception."""
        xknx = XKNX()
        good_callback_1 = Mock()
        bad_callback = Mock(side_effect=Exception("Boom"))
        good_callback_2 = Mock()

        xknx.telegram_queue.register_telegram_received_cb(good_callback_1)
        xknx.telegram_queue.register_telegram_received_cb(bad_callback)
        xknx.telegram_queue.register_telegram_received_cb(good_callback_2)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        await xknx.telegram_queue.process_telegram_incoming(telegram)

        good_callback_1.assert_called_once_with(telegram)
        bad_callback.assert_called_once_with(telegram)
        good_callback_2.assert_called_once_with(telegram)

        logging_exception_mock.assert_called_once_with(
            "Unexpected error while processing telegram_received_cb for %s",
            telegram,
        )

    @pytest.mark.parametrize(
        ("raw_cemi_data_secure", "should_callback_called"),
        [
            # src = 4.0.9; dst = 0/4/0; GroupValueResponse; value=(116, 41, 41)
            # A+C; seq_num=155806854986
            (
                bytes.fromhex("29003ce0400904001103f110002446cfef4ac085e7092ab062b44d"),
                True,  # for GroupValueResponse
            ),
            # Property Value Write PID_GRP_KEY_TABLE connectionless
            # Object Idx = 5, PropId = 35h, Element Count = 1, Index = 1
            # Data = 20 21 22 23 24 25 26 27 28 29 2A 2B 2C 2D 2E 2F
            # A+C
            # from AN158 v07 KNX Data Security AS - Annex A example
            (
                bytes.fromhex(
                    "29 00 b0 60 ff 67 ff 00 22 03 f1 90 00 00 00 00"
                    "00 04 67 67 24 2a 23 08 ca 76 a1 17 74 21 4e e4"
                    "cf 5d 94 90 9f 74 3d 05 0d 8f c1 68"
                ),
                False,  # for destination address is IndividualAddress
            ),
        ],
    )
    def test_data_secure_group_key_issue_callback(
        self, raw_cemi_data_secure: bytes, should_callback_called: bool
    ) -> None:
        """Test incoming undecodable DataSecure Telegram callback."""
        xknx = XKNX()
        xknx.current_address = IndividualAddress("5.0.1")

        data_secure_issue_cb = Mock()
        unsub = xknx.telegram_queue.register_data_secure_group_key_issue_cb(
            data_secure_issue_cb
        )

        xknx.cemi_handler.handle_raw_cemi(raw_cemi_data_secure)
        assert data_secure_issue_cb.called == should_callback_called
        data_secure_issue_cb.reset_mock()

        unsub()
        xknx.cemi_handler.handle_raw_cemi(raw_cemi_data_secure)
        data_secure_issue_cb.assert_not_called()



"""Module for handling values on the KNX bus."""

from .remote_value import GroupAddressesType, RemoteValue
from .remote_value_by_length import RemoteValueByLength
from .remote_value_climate_mode import (
    RemoteValueBinaryHeatCool,
    RemoteValueBinaryOperationMode,
    RemoteValueControllerMode,
    RemoteValueOperationMode,
)
from .remote_value_color_rgb import RemoteValueColorRGB
from .remote_value_color_rgbw import RemoteValueColorRGBW
from .remote_value_color_xyy import RemoteValueColorXYY
from .remote_value_datetime import RemoteValueDate, RemoteValueDateTime, RemoteValueTime
from .remote_value_dpt_value_1_ucount import RemoteValueDptValue1Ucount
from .remote_value_raw import RemoteValueRaw
from .remote_value_scaling import RemoteValueScaling
from .remote_value_scene_number import RemoteValueSceneNumber
from .remote_value_sensor import (
    RemoteValueNumeric,
    RemoteValueSensor,
    RemoteValueString,
)
from .remote_value_setpoint_shift import RemoteValueSetpointShift
from .remote_value_step import RemoteValueStep
from .remote_value_switch import RemoteValueSwitch
from .remote_value_temp import RemoteValueTemp
from .remote_value_updown import RemoteValueUpDown

__all__ = [
    "GroupAddressesType",
    "RemoteValue",
    "RemoteValueBinaryHeatCool",
    "RemoteValueBinaryOperationMode",
    "RemoteValueByLength",
    "RemoteValueColorRGB",
    "RemoteValueColorRGBW",
    "RemoteValueColorXYY",
    "RemoteValueControllerMode",
    "RemoteValueDate",
    "RemoteValueDateTime",
    "RemoteValueDptValue1Ucount",
    "RemoteValueNumeric",
    "RemoteValueOperationMode",
    "RemoteValueRaw",
    "RemoteValueScaling",
    "RemoteValueSceneNumber",
    "RemoteValueSensor",
    "RemoteValueSetpointShift",
    "RemoteValueStep",
    "RemoteValueString",
    "RemoteValueSwitch",
    "RemoteValueTemp",
    "RemoteValueTime",
    "RemoteValueUpDown",
]



"""Module for Profile and resource management ."""

# ruff: noqa: F401
from .const import (
    ResourceDevicePropertyId,
    ResourceEModeChannelPropertyId,
    ResourceEModeDevicePropertyId,
    ResourceGenericPropertyId,
    ResourceGroupObjectTablePropertyId,
    ResourceKNXNETIPPropertyId,
    ResourceLTERoutingTablePropertyId,
    ResourceObjectType,
    ResourcePollingMasterInterfacePropertyId,
    ResourceProgramPropertyId,
    ResourcePropertyId,
    ResourceRFMediumPropertyId,
    ResourceRouterPropertyId,
    ResourceSecureInterfacePropertyId,
    ResourceTextCataloguePropertyId,
)



"""Unit test for value reader."""

import asyncio
from unittest.mock import MagicMock, patch

import pytest

from xknx import XKNX
from xknx.core import ValueReader
from xknx.dpt import DPTBinary
from xknx.telegram import GroupAddress, Telegram, TelegramDirection
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestValueReader:
    """Test class for value reader."""

    async def test_value_reader_read_success(self) -> None:
        """Test value reader: successful read."""
        xknx = XKNX()
        test_group_address = GroupAddress("0/0/1")
        response_telegram = Telegram(
            destination_address=test_group_address,
            direction=TelegramDirection.INCOMING,
            payload=GroupValueResponse(DPTBinary(1)),
        )

        value_reader = ValueReader(xknx, test_group_address)
        # receive the response
        value_reader.telegram_received(response_telegram)
        # and yield the result
        successful_read = await value_reader.read()

        # GroupValueRead telegram is still in the queue because we are not actually processing it
        assert xknx.telegrams.qsize() == 1
        # Callback was removed again
        assert not xknx.telegram_queue.telegram_received_cbs
        # Telegram was received
        assert value_reader.received_telegram == response_telegram
        # successful read() returns the telegram
        assert successful_read == response_telegram

    @patch("logging.Logger.warning")
    async def test_value_reader_read_timeout(
        self, logger_warning_mock: MagicMock
    ) -> None:
        """Test value reader: read timeout."""
        xknx = XKNX()
        value_reader = ValueReader(xknx, GroupAddress("0/0/1"))
        value_reader.response_received_event.wait = MagicMock(
            side_effect=asyncio.TimeoutError()
        )

        timed_out_read = await value_reader.read()

        # GroupValueRead telegram is still in the queue because we are not actually processing it
        assert xknx.telegrams.qsize() == 1
        # Warning was logged
        logger_warning_mock.assert_called_once_with(
            "Error: KNX bus did not respond in time (%s secs) to GroupValueRead request for: %s",
            2.0,
            GroupAddress("0/0/1"),
        )
        # Callback was removed again
        assert not xknx.telegram_queue.telegram_received_cbs
        # No telegram was received
        assert value_reader.received_telegram is None
        # Unsuccessful read() returns None
        assert timed_out_read is None

    async def test_value_reader_read_cancelled(self) -> None:
        """Test value reader: read cancelled."""
        xknx = XKNX()
        value_reader = ValueReader(xknx, GroupAddress("0/0/1"))
        value_reader.response_received_event.wait = MagicMock(
            side_effect=asyncio.CancelledError()
        )
        with pytest.raises(asyncio.CancelledError):
            await value_reader.read()

        # GroupValueRead telegram is still in the queue because we are not actually processing it
        assert xknx.telegrams.qsize() == 1
        # Callback was removed again
        assert not xknx.telegram_queue.telegram_received_cbs
        # No telegram was received
        assert value_reader.received_telegram is None

    async def test_value_reader_send_group_read(self) -> None:
        """Test value reader: send_group_read."""
        xknx = XKNX()
        value_reader = ValueReader(xknx, GroupAddress("0/0/1"))

        value_reader.send_group_read()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("0/0/1"), payload=GroupValueRead()
        )

    async def test_value_reader_telegram_received(self) -> None:
        """Test value reader: telegram_received."""
        xknx = XKNX()
        test_group_address = GroupAddress("0/0/1")
        expected_telegram_1 = Telegram(
            destination_address=test_group_address,
            direction=TelegramDirection.INCOMING,
            payload=GroupValueResponse(DPTBinary(1)),
        )
        expected_telegram_2 = Telegram(
            destination_address=test_group_address,
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        telegram_wrong_address = Telegram(
            destination_address=GroupAddress("0/0/2"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueResponse(DPTBinary(1)),
        )
        telegram_wrong_type = Telegram(
            destination_address=test_group_address,
            direction=TelegramDirection.INCOMING,
            payload=GroupValueRead(),
        )

        value_reader = ValueReader(xknx, test_group_address)

        value_reader.telegram_received(telegram_wrong_address)
        assert value_reader.received_telegram is None
        assert not value_reader.response_received_event.is_set()

        value_reader.telegram_received(telegram_wrong_type)
        assert value_reader.received_telegram is None
        assert not value_reader.response_received_event.is_set()

        value_reader.telegram_received(expected_telegram_1)
        assert value_reader.received_telegram == expected_telegram_1
        assert value_reader.response_received_event.is_set()

        value_reader.telegram_received(expected_telegram_2)
        assert value_reader.received_telegram == expected_telegram_2
        assert value_reader.response_received_event.is_set()



"""
Module for reading the value of a specific KNX group address from KNX bus.

The module will
* ... send a group_read to the selected group address.
* ... register a callback for receiving telegrams within telegram queue.
* ... check if received telegrams have the correct group address.
* ... store the received telegram for further processing.
"""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

from xknx.telegram import Telegram
from xknx.telegram.address import GroupAddress, InternalGroupAddress
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite
from xknx.util import asyncio_timeout

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class ValueReader:
    """Class for reading the value of a specific KNX group address from KNX bus."""

    __slots__ = (
        "group_address",
        "received_telegram",
        "response_received_event",
        "timeout_in_seconds",
        "xknx",
    )

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddress | InternalGroupAddress,
        timeout_in_seconds: float = 2.0,
    ) -> None:
        """Initialize ValueReader class."""
        self.xknx = xknx
        self.group_address: GroupAddress | InternalGroupAddress = group_address
        self.response_received_event = asyncio.Event()
        self.timeout_in_seconds: float = timeout_in_seconds
        self.received_telegram: Telegram | None = None

    async def read(self) -> Telegram | None:
        """Send group read and wait for response."""
        cb_obj = self.xknx.telegram_queue.register_telegram_received_cb(
            self.telegram_received,
            group_addresses=[self.group_address],
            match_for_outgoing=True,
        )
        self.send_group_read()

        try:
            async with asyncio_timeout(self.timeout_in_seconds):
                await self.response_received_event.wait()
        except asyncio.TimeoutError:
            logger.warning(
                "Error: KNX bus did not respond in time (%s secs) to GroupValueRead request for: %s",
                self.timeout_in_seconds,
                self.group_address,
            )
        else:
            return self.received_telegram
        finally:
            # cleanup to not leave callbacks (for asyncio.CancelledError)
            self.xknx.telegram_queue.unregister_telegram_received_cb(cb_obj)

        return None

    def send_group_read(self) -> None:
        """Send group read."""
        telegram = Telegram(
            destination_address=self.group_address,
            payload=GroupValueRead(),
            source_address=self.xknx.current_address,
        )
        self.xknx.telegrams.put_nowait(telegram)

    def telegram_received(self, telegram: Telegram) -> None:
        """Test if telegram has correct group address and trigger event."""
        if telegram.destination_address == self.group_address and isinstance(
            telegram.payload, GroupValueResponse | GroupValueWrite
        ):
            self.received_telegram = telegram
            self.response_received_event.set()



"""Utilities for KNX Secure."""

from cryptography.hazmat.primitives import hashes

# py3.12 collections.abc.Buffer can be used instead
Buffer = bytes | bytearray | memoryview


def bytes_xor(a: bytes, b: bytes) -> bytes:  # pylint: disable=invalid-name
    """
    XOR two bytes values.

    Different lengths raise ValueError.
    """
    if len(a) != len(b):
        raise ValueError("Length of a and b must be equal.")
    return (int.from_bytes(a, "big") ^ int.from_bytes(b, "big")).to_bytes(len(a), "big")


def byte_pad(data: bytes, block_size: int) -> bytes:
    """Pad data with 0x00 until its length is a multiple of block_size."""
    if remainder := len(data) % block_size:
        return data + bytes(block_size - remainder)
    return data


def sha256_hash(data: Buffer) -> bytes:
    """Calculate SHA256 hash of data."""
    digest = hashes.Hash(hashes.SHA256())
    digest.update(data)
    return digest.finalize()



"""Example for switching a light on and off."""

import asyncio

from xknx import XKNX
from xknx.devices import Light


async def main() -> None:
    """Connect to KNX/IP bus, slowly dimm on light, set it off again afterwards."""
    xknx = XKNX()
    await xknx.start()

    light = Light(
        xknx,
        name="TestLight2",
        group_address_switch="1/0/12",
        group_address_brightness="1/0/14",
    )
    xknx.devices.async_add(light)

    for i in [0, 31, 63, 95, 127, 159, 191, 223, 255]:
        await light.set_brightness(i)
        await asyncio.sleep(1)

    await light.set_off()

    await xknx.stop()


asyncio.run(main())



"""Unit test for KNX DPT 235 objects."""

from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTTariffActiveEnergy, TariffActiveEnergy
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestTariffActiveEnergy:
    """Test TariffActiveEnergy class."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            (
                {"energy": -2_147_483_648, "tariff": 0},
                TariffActiveEnergy(-2_147_483_648, 0),
            ),
            (
                {"energy": 2_147_483_647, "tariff": 255},
                TariffActiveEnergy(2_147_483_647, 255),
            ),
            ({"tariff": 128}, TariffActiveEnergy(None, 128)),
            ({"energy": 50}, TariffActiveEnergy(50, None)),
            ({}, TariffActiveEnergy()),
            (
                {"energy": 255, "tariff": None},
                TariffActiveEnergy(255, None),
            ),
            (
                {"energy": None, "tariff": 128},
                TariffActiveEnergy(None, 128),
            ),
            ({"energy": 255, "tariff": None}, TariffActiveEnergy(255, None)),
            ({"tariff": 128}, TariffActiveEnergy(None, 128)),
            ({"energy": 128}, TariffActiveEnergy(128, None)),
            ({"energy": 128, "tariff": 128}, TariffActiveEnergy(128, 128)),
        ],
    )
    def test_dict(self, data: dict[str, Any], value: TariffActiveEnergy) -> None:
        """Test from_dict and as_dict methods."""
        test_value = TariffActiveEnergy.from_dict(data)
        assert test_value == value
        # fields default to `None`
        default_dict = {"energy": None, "tariff": None}
        assert value.as_dict() == default_dict | data

    @pytest.mark.parametrize(
        "data",
        [
            # invalid data
            {"energy": 128, "tariff": "a"},
            {"energy": "a", "tariff": 128},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            TariffActiveEnergy.from_dict(data)


class TestDPTTariffActiveEnergy:
    """Test class for KNX DPTTariffActiveEnergy objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (
                TariffActiveEnergy(0, 0),
                (0x00, 0x00, 0x00, 0x00, 0x00, 0x00),
            ),  # zero values
            (
                TariffActiveEnergy(-2_147_483_648, 0),
                (0x80, 0x00, 0x00, 0x00, 0x00, 0x00),
            ),  # min values
            (
                TariffActiveEnergy(2_147_483_647, 254),
                (0x7F, 0xFF, 0xFF, 0xFF, 0xFE, 0x00),
            ),  # max values
            (
                TariffActiveEnergy(None, 0),
                (0x00, 0x00, 0x00, 0x00, 0x00, 0x02),
            ),  # energy None
            (
                TariffActiveEnergy(0, None),
                (0x00, 0x00, 0x00, 0x00, 0x00, 0x01),
            ),  # tariff None
            (TariffActiveEnergy(None, None), (0x00, 0x00, 0x00, 0x00, 0x00, 0x03)),
            (TariffActiveEnergy(128, 128), (0x00, 0x00, 0x00, 0x80, 0x80, 0x00)),
            (TariffActiveEnergy(204, 204), (0x00, 0x00, 0x00, 0xCC, 0xCC, 0x00)),
        ],
    )
    def test_dpt_tariff_active_energy_value(
        self, value: TariffActiveEnergy, raw: tuple[int, ...]
    ) -> None:
        """Test DPTTariffActiveEnergy parsing and streaming."""
        knx_value = DPTTariffActiveEnergy.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTTariffActiveEnergy.from_knx(knx_value) == value

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            ({"tariff": 128, "energy": 128}, (0x00, 0x00, 0x00, 0x80, 0x80, 0x00)),
            ({"tariff": 204, "energy": 204}, (0x00, 0x00, 0x00, 0xCC, 0xCC, 0x00)),
        ],
    )
    def test_dpt_tariff_active_energy_to_knx_from_dict(
        self, value: dict[str, Any], raw: tuple[int, ...]
    ) -> None:
        """Test DPTTariffActiveEnergy parsing from a dict."""
        knx_value = DPTTariffActiveEnergy.to_knx(value)
        assert knx_value == DPTArray(raw)

    @pytest.mark.parametrize(
        ("value"),
        [
            TariffActiveEnergy(-2_147_483_649, 0),
            TariffActiveEnergy(2_147_483_648, 0),
            TariffActiveEnergy(0, -1),
            TariffActiveEnergy(0, 255),
        ],
    )
    def test_dpt_tariff_active_energy_to_knx_limits(
        self, value: TariffActiveEnergy
    ) -> None:
        """Test initialization of DPTTariffActiveEnergy with wrong value."""
        with pytest.raises(ConversionError):
            DPTTariffActiveEnergy.to_knx(value)

    @pytest.mark.parametrize(
        "value",
        [
            None,
            (0xFF, 0x4E),
            1,
            ((0x00, 0xFF), 0x4E),
            ((0xFF, 0x4E), (0x12, 0x00)),
            ((0, 0), "a"),
            ((0, 0), 0.4),
            TariffActiveEnergy(tariff=0, energy="a"),
            TariffActiveEnergy(tariff="a", energy=0),
        ],
    )
    def test_dpt_tariff_active_energy_wrong_value_to_knx(self, value: Any) -> None:
        """Test DPTTariffActiveEnergy parsing with wrong value."""
        with pytest.raises(ConversionError):
            DPTTariffActiveEnergy.to_knx(value)

    def test_dpt_tariff_active_energy_wrong_value_from_knx(self) -> None:
        """Test DPTTariffActiveEnergy parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTTariffActiveEnergy.from_knx(DPTArray((0xFF, 0x4E)))



"""Module for KNX/IP Error codes."""

from enum import Enum


class ErrorCode(Enum):
    """Enum class for KNX/IP error codes."""

    # The connection state is normal.
    E_NO_ERROR = 0x00

    # requested host protocol is not supported
    E_HOST_PROTOCOL_TYPE = 0x01

    # requested protocol version is not supported
    E_VERSION_NOT_SUPPORTED = 0x02

    # received sequence number is out of order.
    E_SEQUENCE_NUMBER = 0x04

    # Any further undefined, possibly implementation specific error has occurred.
    # Core v2
    E_ERROR = 0x0F

    # The KNXnet/IP Server device cannot find an active data
    # connection with the specified ID.
    E_CONNECTION_ID = 0x21

    # The requested connection type is not supported
    E_CONNECTION_TYPE = 0x22

    # One or more requested connection options are not supported
    E_CONNECTION_OPTION = 0x23

    # The KNXnet/IP Server device cannot accept the new data connection
    # because its maximum amount of concurrent connections is already
    # occupied.
    E_NO_MORE_CONNECTIONS = 0x24

    # KNXnet/IP Tunnelling device does not accept connection because the
    # Individual Address is used multiple times
    E_NO_MORE_UNIQUE_CONNECTIONS = 0x25

    # The KNXnet/IP Server device detects an error concerning
    # the data connection with the specified ID.
    E_DATA_CONNECTION = 0x26

    # The KNXnet/IP Server device detects an error concerning
    # the KNX subnetwork connection with the specified ID.
    E_KNX_CONNECTION = 0x27

    # The Client is not authorised to use the requested IA in the Extended CRI.
    # Core v2
    E_AUTHORISATION_ERROR = 0x28

    # The requested tunnelling layer is not supported by the
    # KNXnet/IP Server device.
    E_TUNNELLING_LAYER = 0x29

    # The IA requested in the Extended CRI is not a Tunnelling IA.
    # Core v2
    E_NO_TUNNELLING_ADDRESS = 0x2D

    # The IA requested for this connection is in use.
    # Core v2
    E_CONNECTION_IN_USE = 0x2E



"""Unit test for Sensor objects."""

from unittest.mock import Mock

import pytest

from xknx import XKNX
from xknx.devices import Sensor
from xknx.dpt import DPTArray
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestSensor:
    """Test class for Sensor objects."""

    @pytest.mark.parametrize(
        ("value_type", "raw_payload", "expected_state"),
        [
            # DPT-14 values are according to ETS group monitor values
            (
                "absolute_temperature",
                DPTArray((0x44, 0xD7, 0xD2, 0x8B)),
                1726.579,
            ),
            (
                "acceleration",
                DPTArray((0x45, 0x94, 0xD8, 0x5D)),
                4763.045,
            ),
            (
                "volume_liquid_litre",
                DPTArray((0x00, 0x00, 0x01, 0x00)),
                256,
            ),
            (
                "volume_m3",
                DPTArray((0x00, 0x00, 0x01, 0x00)),
                256,
            ),
            (
                "active_energy",
                DPTArray((0x26, 0x37, 0x49, 0x7F)),
                641157503,
            ),
            (
                "active_energy_kwh",
                DPTArray((0x37, 0x5, 0x5, 0xEA)),
                923076074,
            ),
            (
                "activity",
                DPTArray((0x45, 0x76, 0x0, 0xA3)),
                3936.04,
            ),
            (
                "amplitude",
                DPTArray((0x45, 0x9A, 0xED, 0x8)),
                4957.629,
            ),
            (
                "angle",
                DPTArray((0xE4,)),
                322,
            ),
            (
                "angle_deg",
                DPTArray((0x44, 0x5C, 0x20, 0x2B)),
                880.5026,
            ),
            (
                "angle_rad",
                DPTArray((0x44, 0x36, 0x75, 0x1)),
                729.8282,
            ),
            (
                "angular_frequency",
                DPTArray((0x43, 0xBC, 0x20, 0x8D)),
                376.2543,
            ),
            (
                "angular_momentum",
                DPTArray((0xC2, 0x75, 0xB7, 0xB5)),
                -61.4294,
            ),
            (
                "angular_velocity",
                DPTArray((0xC4, 0xD9, 0x10, 0xB3)),
                -1736.522,
            ),
            (
                "apparant_energy",
                DPTArray((0xD3, 0xBD, 0x1E, 0xA5)),
                -742580571,
            ),
            (
                "apparant_energy_kvah",
                DPTArray((0x49, 0x40, 0xC9, 0x9)),
                1228982537,
            ),
            (
                "area",
                DPTArray((0x45, 0x63, 0x1E, 0xCD)),
                3633.925,
            ),
            (
                "brightness",
                DPTArray((0xC3, 0x56)),
                50006,
            ),
            (
                "capacitance",
                DPTArray((0x45, 0xC9, 0x1D, 0x9D)),
                6435.702,
            ),
            (
                "charge_density_surface",
                DPTArray((0x45, 0xDB, 0x66, 0x99)),
                7020.825,
            ),
            (
                "charge_density_volume",
                DPTArray((0xC4, 0x8C, 0x33, 0xD7)),
                -1121.62,
            ),
            (
                "color_temperature",
                DPTArray((0x6C, 0x95)),
                27797,
            ),
            (
                "common_temperature",
                DPTArray((0x45, 0xD9, 0xC6, 0x3F)),
                6968.781,
            ),
            (
                "compressibility",
                DPTArray((0x45, 0x89, 0x94, 0xAB)),
                4402.583,
            ),
            (
                "conductance",
                DPTArray((0x45, 0xA6, 0x28, 0xF9)),
                5317.122,
            ),
            (
                "counter_pulses",
                DPTArray((0x9D,)),
                -99,
            ),
            (
                "current",
                DPTArray((0xCA, 0xCC)),
                51916,
            ),
            (
                "delta_time_hrs",
                DPTArray((0x47, 0x80)),
                18304,
            ),
            (
                "delta_time_min",
                DPTArray((0xB9, 0x7B)),
                -18053,
            ),
            (
                "delta_time_ms",
                DPTArray((0x58, 0x77)),
                22647,
            ),
            (
                "delta_time_sec",
                DPTArray((0xA3, 0x6A)),
                -23702,
            ),
            (
                "density",
                DPTArray((0x44, 0xA5, 0xCB, 0x27)),
                1326.349,
            ),
            (
                "electrical_conductivity",
                DPTArray((0xC4, 0xC6, 0xF5, 0x6E)),
                -1591.67,
            ),
            (
                "electric_charge",
                DPTArray((0x46, 0x14, 0xF6, 0xA0)),
                9533.656,
            ),
            (
                "electric_current",
                DPTArray((0x45, 0xAD, 0x45, 0x90)),
                5544.695,
            ),
            (
                "electric_current_density",
                DPTArray((0x45, 0x7C, 0x57, 0xF6)),
                4037.498,
            ),
            (
                "electric_dipole_moment",
                DPTArray((0x45, 0x58, 0xF1, 0x73)),
                3471.091,
            ),
            (
                "electric_displacement",
                DPTArray((0xC5, 0x34, 0x8B, 0x0)),
                -2888.688,
            ),
            (
                "electric_field_strength",
                DPTArray((0xC6, 0x17, 0x1C, 0x39)),
                -9671.056,
            ),
            (
                "electric_flux",
                DPTArray((0x45, 0x8F, 0x6C, 0xFD)),
                4589.624,
            ),
            (
                "electric_flux_density",
                DPTArray((0xC6, 0x0, 0x50, 0xA8)),
                -8212.164,
            ),
            (
                "electric_polarization",
                DPTArray((0x45, 0xF8, 0x89, 0xC6)),
                7953.222,
            ),
            (
                "electric_potential",
                DPTArray((0xC6, 0x18, 0xA4, 0xAF)),
                -9769.171,
            ),
            (
                "electric_potential_difference",
                DPTArray((0xC6, 0xF, 0x1D, 0x6)),
                -9159.256,
            ),
            (
                "electromagnetic_moment",
                DPTArray((0x45, 0x82, 0x48, 0xAE)),
                4169.085,
            ),
            (
                "electromotive_force",
                DPTArray((0x45, 0xBC, 0xEF, 0xEB)),
                6045.99,
            ),
            (
                "energy",
                DPTArray((0x45, 0x4B, 0xB3, 0xF8)),
                3259.248,
            ),
            (
                "enthalpy",
                DPTArray((0x76, 0xDD)),
                287866.88,
            ),
            (
                "flow_rate_m3h",
                DPTArray((0x99, 0xEA, 0xC0, 0x55)),
                -1712668587,
            ),
            (
                "force",
                DPTArray((0x45, 0x9E, 0x2C, 0xE1)),
                5061.61,
            ),
            (
                "frequency",
                DPTArray((0x45, 0xC2, 0x3C, 0x44)),
                6215.533,
            ),
            (
                "heatcapacity",
                DPTArray((0xC5, 0xB3, 0x56, 0x7E)),
                -5738.812,
            ),
            (
                "heatflowrate",
                DPTArray((0x44, 0xEC, 0x80, 0x7A)),
                1892.015,
            ),
            (
                "heat_quantity",
                DPTArray((0xC5, 0xA6, 0xB6, 0xD5)),
                -5334.854,
            ),
            (
                "humidity",
                DPTArray((0x7E, 0xE1)),
                577044.48,
            ),
            (
                "impedance",
                DPTArray((0x45, 0xDD, 0x79, 0x6D)),
                7087.178,
            ),
            (
                "illuminance",
                DPTArray((0x7C, 0x5E)),
                366346.24,
            ),
            (
                "kelvin_per_percent",
                DPTArray((0xFA, 0xBD)),
                -441384.96,
            ),
            (
                "length",
                DPTArray((0xC5, 0x9D, 0xAE, 0xC5)),
                -5045.846,
            ),
            (
                "length_mm",
                DPTArray((0x56, 0xB9)),
                22201,
            ),
            (
                "light_quantity",
                DPTArray((0x45, 0x4A, 0xF5, 0x68)),
                3247.338,
            ),
            (
                "long_delta_timesec",
                DPTArray((0x45, 0xB2, 0x17, 0x54)),
                1169299284,
            ),
            (
                "luminance",
                DPTArray((0x45, 0x18, 0xD9, 0x76)),
                2445.591,
            ),
            (
                "luminous_flux",
                DPTArray((0x45, 0xBD, 0x16, 0x9)),
                6050.754,
            ),
            (
                "luminous_intensity",
                DPTArray((0x46, 0xB, 0xBE, 0x7E)),
                8943.623,
            ),
            (
                "magnetic_field_strength",
                DPTArray((0x44, 0x15, 0xF1, 0xAD)),
                599.7762,
            ),
            (
                "magnetic_flux",
                DPTArray((0xC5, 0xCB, 0x3C, 0x98)),
                -6503.574,
            ),
            (
                "magnetic_flux_density",
                DPTArray((0x45, 0xB6, 0xBD, 0x42)),
                5847.657,
            ),
            (
                "magnetic_moment",
                DPTArray((0xC3, 0x8E, 0x7F, 0x73)),
                -284.9957,
            ),
            (
                "magnetic_polarization",
                DPTArray((0x45, 0x8C, 0xFA, 0xCB)),
                4511.349,
            ),
            (
                "magnetization",
                DPTArray((0x45, 0xF7, 0x9D, 0xA2)),
                7923.704,
            ),
            (
                "magnetomotive_force",
                DPTArray((0xC6, 0x4, 0xC2, 0xDA)),
                -8496.713,
            ),
            (
                "mass",
                DPTArray((0x45, 0x8F, 0x70, 0xA4)),
                4590.08,
            ),
            (
                "mass_flux",
                DPTArray((0xC6, 0x7, 0x34, 0xFF)),
                -8653.249,
            ),
            (
                "mol",
                DPTArray((0xC4, 0xA0, 0xF4, 0x68)),
                -1287.638,
            ),
            (
                "momentum",
                DPTArray((0xC5, 0x27, 0xAA, 0x5B)),
                -2682.647,
            ),
            (
                "percent",
                DPTArray((0xE3,)),
                89,
            ),
            (
                "percentU8",
                DPTArray((0x6B,)),
                107,
            ),
            (
                "percentV8",
                DPTArray((0x20,)),
                32,
            ),
            (
                "percentV16",
                DPTArray((0x8A, 0x2F)),
                -301.61,
            ),
            (
                "phaseanglerad",
                DPTArray((0x45, 0x54, 0xAC, 0x2E)),
                3402.761,
            ),
            (
                "phaseangledeg",
                DPTArray((0xC5, 0x25, 0x13, 0x38)),
                -2641.201,
            ),
            (
                "power",
                DPTArray((0x45, 0xCB, 0xE2, 0x5C)),
                6524.295,
            ),
            (
                "power_2byte",
                DPTArray((0x6D, 0x91)),
                116736.0,
            ),
            (
                "power_density",
                DPTArray((0x65, 0x3E)),
                54968.32,
            ),
            (
                "powerfactor",
                DPTArray((0xC5, 0x35, 0x28, 0x21)),
                -2898.508,
            ),
            (
                "ppm",
                DPTArray((0x7F, 0x74)),
                625213.44,
            ),
            (
                "pressure",
                DPTArray((0xC5, 0xE6, 0xE6, 0x63)),
                -7388.798,
            ),
            (
                "pressure_2byte",
                DPTArray((0x7C, 0xF4)),
                415498.24,
            ),
            (
                "pulse",
                DPTArray((0xFC,)),
                252,
            ),
            (
                "rain_amount",
                DPTArray((0xE0, 0xD0)),
                -75366.4,
            ),
            (
                "reactance",
                DPTArray((0x45, 0xB0, 0x50, 0x91)),
                5642.071,
            ),
            (
                "reactive_energy",
                DPTArray((0x1A, 0x49, 0x6D, 0xA7)),
                441019815,
            ),
            (
                "reactive_energy_kvarh",
                DPTArray((0xCC, 0x62, 0x5, 0x31)),
                -865991375,
            ),
            (
                "resistance",
                DPTArray((0xC5, 0xFC, 0x5F, 0xC2)),
                -8075.97,
            ),
            (
                "resistivity",
                DPTArray((0xC5, 0x57, 0x76, 0xC3)),
                -3447.423,
            ),
            (
                "rotation_angle",
                DPTArray((0x2D, 0xDC)),
                11740,
            ),
            (
                "scene_number",
                DPTArray((0x1,)),
                2,
            ),
            (
                "self_inductance",
                DPTArray((0xC4, 0xA1, 0xB0, 0x6)),
                -1293.501,
            ),
            (
                "solid_angle",
                DPTArray((0xC5, 0xC6, 0xE5, 0x47)),
                -6364.66,
            ),
            (
                "sound_intensity",
                DPTArray((0xC4, 0xF2, 0x56, 0xE6)),
                -1938.716,
            ),
            (
                "speed",
                DPTArray((0xC5, 0xCD, 0x1C, 0x6A)),
                -6563.552,
            ),
            (
                "stress",
                DPTArray((0x45, 0xDC, 0xA8, 0xF2)),
                7061.118,
            ),
            (
                "surface_tension",
                DPTArray((0x46, 0xB, 0xAC, 0x11)),
                8939.017,
            ),
            (
                "temperature",
                DPTArray((0x77, 0x88)),
                315883.52,
            ),
            (
                "temperature_a",
                DPTArray((0xF1, 0xDB)),
                -257720.32,
            ),
            (
                "temperature_difference",
                DPTArray((0xC6, 0xC, 0x50, 0xBC)),
                -8980.184,
            ),
            (
                "temperature_difference_2byte",
                DPTArray((0xA9, 0xF4)),
                -495.36,
            ),
            (
                "temperature_f",
                DPTArray((0x67, 0xA9)),
                80322.56,
            ),
            (
                "thermal_capacity",
                DPTArray((0x45, 0x83, 0xEA, 0xB3)),
                4221.337,
            ),
            (
                "thermal_conductivity",
                DPTArray((0xC5, 0x9C, 0x4D, 0x22)),
                -5001.642,
            ),
            (
                "thermoelectric_power",
                DPTArray((0x41, 0xCF, 0x9E, 0x4F)),
                25.9523,
            ),
            (
                "time_1",
                DPTArray((0x5E, 0x1E)),
                32071.68,
            ),
            (
                "time_2",
                DPTArray((0xFB, 0x29)),
                -405995.52,
            ),
            (
                "time_period_100msec",
                DPTArray((0x6A, 0x35)),
                2718900,
            ),
            (
                "time_period_10msec",
                DPTArray((0x32, 0x3)),
                128030,
            ),
            (
                "time_period_hrs",
                DPTArray((0x29, 0xDE)),
                10718,
            ),
            (
                "time_period_min",
                DPTArray((0x0, 0x54)),
                84,
            ),
            (
                "time_period_msec",
                DPTArray((0x93, 0xC7)),
                37831,
            ),
            (
                "time_period_sec",
                DPTArray((0xE0, 0xF5)),
                57589,
            ),
            (
                "time_seconds",
                DPTArray((0x45, 0xEC, 0x91, 0x7C)),
                7570.186,
            ),
            (
                "torque",
                DPTArray((0xC5, 0x9, 0x23, 0x5F)),
                -2194.211,
            ),
            (
                "voltage",
                DPTArray((0x6D, 0xBF)),
                120504.32,
            ),
            (
                "volume",
                DPTArray((0x46, 0x16, 0x98, 0x43)),
                9638.065,
            ),
            (
                "volume_flow",
                DPTArray((0x7C, 0xF5)),
                415825.92,
            ),
            (
                "volume_flux",
                DPTArray((0xC5, 0x4, 0x2D, 0x72)),
                -2114.84,
            ),
            (
                "weight",
                DPTArray((0x45, 0x20, 0x10, 0xE8)),
                2561.057,
            ),
            (
                "work",
                DPTArray((0x45, 0x64, 0x5D, 0xBE)),
                3653.859,
            ),
            (
                "wind_speed_ms",
                DPTArray((0x7D, 0x98)),
                469237.76,
            ),
            (
                "wind_speed_kmh",
                DPTArray((0x68, 0x0)),
                0.0,
            ),
        ],
    )
    async def test_sensor_value_types(
        self,
        value_type: str,
        raw_payload: DPTArray,
        expected_state: float,
    ) -> None:
        """Test sensor value types."""
        xknx = XKNX()
        sensor = Sensor(
            xknx,
            "TestSensor",
            group_address_state="1/2/3",
            value_type=value_type,
        )
        sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(value=raw_payload),
            )
        )

        assert sensor.resolve_state() == expected_state

    async def test_always_callback_sensor(self) -> None:
        """Test always callback sensor."""
        xknx = XKNX()
        sensor = Sensor(
            xknx,
            "TestSensor",
            group_address_state="1/2/3",
            always_callback=False,
            value_type="volume_liquid_litre",
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)
        payload = DPTArray((0x00, 0x00, 0x01, 0x00))
        #  set initial payload of sensor
        sensor.sensor_value.value = 256
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueWrite(payload)
        )
        response_telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(payload),
        )
        # verify not called when always_callback is False
        sensor.process(telegram)
        after_update_callback.assert_not_called()
        after_update_callback.reset_mock()

        # verify not called when processing read responses
        sensor.process(response_telegram)
        after_update_callback.assert_not_called()
        after_update_callback.reset_mock()

        sensor.always_callback = True
        # verify called when always_callback is True
        sensor.process(telegram)
        after_update_callback.assert_called_once()
        after_update_callback.reset_mock()

        # verify called when processing read responses
        sensor.process(response_telegram)
        after_update_callback.assert_called_once()

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        sensor = Sensor(
            xknx, "TestSensor", value_type="temperature", group_address_state="1/2/3"
        )
        await sensor.sync()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )

    #
    # HAS GROUP ADDRESS
    #
    def test_has_group_address(self) -> None:
        """Test sensor has group address."""
        xknx = XKNX()
        sensor = Sensor(
            xknx, "TestSensor", value_type="temperature", group_address_state="1/2/3"
        )
        assert sensor.has_group_address(GroupAddress("1/2/3"))
        assert not sensor.has_group_address(GroupAddress("1/2/4"))

    #
    # TEST PROCESS
    #
    async def test_process(self) -> None:
        """Test process / reading telegrams from telegram queue."""
        xknx = XKNX()
        sensor = Sensor(
            xknx, "TestSensor", value_type="temperature", group_address_state="1/2/3"
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x06, 0xA0))),
        )
        sensor.process(telegram)
        assert sensor.sensor_value.value == 16.96
        assert sensor.sensor_value.telegram.payload.value == DPTArray((0x06, 0xA0))
        assert sensor.resolve_state() == 16.96
        # test HomeAssistant device class
        assert sensor.ha_device_class() == "temperature"

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        sensor = Sensor(
            xknx, "TestSensor", group_address_state="1/2/3", value_type="temperature"
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01, 0x02))),
        )
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)
        assert sensor.last_telegram == telegram



"""Module for the automations and business logic of XKNX."""

# ruff: noqa: F401
from .connection_manager import ConnectionManager
from .connection_state import XknxConnectionState, XknxConnectionType
from .group_address_dpt import GroupAddressDPT
from .state_updater import StateUpdater
from .task_registry import Task, TaskRegistry
from .telegram_queue import TelegramQueue
from .value_reader import ValueReader



"""Test management handling."""

import asyncio
from unittest.mock import AsyncMock, call, patch

import pytest

from xknx import XKNX
from xknx.exceptions import (
    CommunicationError,
    ConfirmationError,
    ManagementConnectionError,
    ManagementConnectionTimeout,
)
from xknx.management.management import MANAGAMENT_ACK_TIMEOUT
from xknx.telegram import (
    GroupAddress,
    IndividualAddress,
    Telegram,
    TelegramDirection,
    apci,
    tpci,
)

from ..conftest import EventLoopClockAdvancer


async def test_connect() -> None:
    """Test establishing connections."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    ia_1 = IndividualAddress("4.0.1")
    ia_2 = IndividualAddress("4.0.2")

    def tg_connect(ia: IndividualAddress) -> Telegram:
        return Telegram(
            source_address=xknx.current_address,
            destination_address=ia,
            direction=TelegramDirection.OUTGOING,
            tpci=tpci.TConnect(),
        )

    def tg_disconnect(ia: IndividualAddress) -> Telegram:
        return Telegram(
            source_address=xknx.current_address,
            destination_address=ia,
            direction=TelegramDirection.OUTGOING,
            tpci=tpci.TDisconnect(),
        )

    await xknx.management.connect(ia_1)
    conn_2 = await xknx.management.connect(ia_2)

    with pytest.raises(ManagementConnectionError):
        # no 2 connections to the same IA
        await xknx.management.connect(ia_1)

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(tg_connect(ia_1)),
        call(tg_connect(ia_2)),
    ]
    xknx.cemi_handler.send_telegram.reset_mock()

    await xknx.management.disconnect(ia_1)
    await conn_2.disconnect()

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(tg_disconnect(ia_1)),
        call(tg_disconnect(ia_2)),
    ]

    # connect again doesn't raise
    await xknx.management.connect(ia_1)


async def test_ack_timeout(time_travel: EventLoopClockAdvancer) -> None:
    """Test ACK timeout handling."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    _ia = IndividualAddress("4.0.1")

    conn = await xknx.management.connect(_ia)
    xknx.cemi_handler.send_telegram.reset_mock()

    device_desc_read = Telegram(
        destination_address=_ia,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    task = asyncio.create_task(
        conn.request(
            payload=apci.DeviceDescriptorRead(descriptor=0),
            expected=apci.DeviceDescriptorResponse,
        )
    )
    await asyncio.sleep(0)
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(device_desc_read),
    ]
    await time_travel(MANAGAMENT_ACK_TIMEOUT)
    # telegram repeated
    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(device_desc_read),
        call(device_desc_read),
    ]
    await time_travel(MANAGAMENT_ACK_TIMEOUT)
    with pytest.raises(ManagementConnectionTimeout):
        # still no ACK -> timeout
        await task

    await conn.disconnect()


async def test_failed_connect_disconnect() -> None:
    """Test failing connections."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    ia_1 = IndividualAddress("4.0.1")

    xknx.cemi_handler.send_telegram.side_effect = ConfirmationError("")
    with pytest.raises(ManagementConnectionError):
        await xknx.management.connect(ia_1)

    xknx.cemi_handler.send_telegram.side_effect = CommunicationError("")
    with pytest.raises(ManagementConnectionError):
        await xknx.management.connect(ia_1)

    xknx.cemi_handler.send_telegram.side_effect = None
    conn_1 = await xknx.management.connect(ia_1)
    xknx.cemi_handler.send_telegram.side_effect = ConfirmationError("")
    with pytest.raises(ManagementConnectionError):
        await xknx.management.disconnect(ia_1)

    xknx.cemi_handler.send_telegram.side_effect = None
    conn_1 = await xknx.management.connect(ia_1)
    xknx.cemi_handler.send_telegram.side_effect = CommunicationError("")
    with pytest.raises(ManagementConnectionError):
        await conn_1.disconnect()


async def test_reject_incoming_connection() -> None:
    """Test rejecting incoming transport connections."""
    # Note: incoming L_DATA.ind indication connection requests are rejected
    # L_DATA.req frames received from a tunnelling client are not yet supported
    xknx = XKNX()
    individual_address = IndividualAddress("4.0.10")

    connect = Telegram(
        source_address=individual_address,
        destination_address=xknx.current_address,
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TConnect(),
    )
    disconnect = Telegram(
        source_address=xknx.current_address,
        destination_address=individual_address,
        tpci=tpci.TDisconnect(),
    )
    with patch("xknx.cemi.CEMIHandler.send_telegram") as send_telegram:
        xknx.cemi_handler.telegram_received(connect)
        await asyncio.sleep(0)
        assert send_telegram.call_args_list == [call(disconnect)]


async def test_incoming_unexpected_numbered_telegram() -> None:
    """Test incoming unexpected numbered telegram is acked."""
    xknx = XKNX()
    individual_address = IndividualAddress("4.0.10")

    device_desc_read = Telegram(
        source_address=individual_address,
        destination_address=xknx.current_address,
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDataConnected(0),
        payload=apci.DeviceDescriptorRead(descriptor=0),
    )
    ack = Telegram(
        source_address=xknx.current_address,
        destination_address=individual_address,
        direction=TelegramDirection.OUTGOING,
        tpci=tpci.TAck(0),
    )
    with patch("xknx.cemi.CEMIHandler.send_telegram") as send_telegram:
        xknx.cemi_handler.telegram_received(device_desc_read)
        await asyncio.sleep(0)
        assert send_telegram.call_args_list == [call(ack)]


async def test_incoming_wrong_address() -> None:
    """Test incoming telegrams addressed to different devices."""
    xknx = XKNX()
    individual_address = IndividualAddress("4.0.10")
    other_address = IndividualAddress("4.0.11")
    assert xknx.current_address != other_address

    connect = Telegram(
        source_address=individual_address,
        destination_address=other_address,
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TConnect(),
    )
    ack = Telegram(
        source_address=individual_address,
        destination_address=other_address,
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TAck(0),
    )
    disconnect = Telegram(
        source_address=individual_address,
        destination_address=other_address,
        direction=TelegramDirection.INCOMING,
        tpci=tpci.TDisconnect(),
    )
    with patch("xknx.cemi.CEMIHandler.send_telegram") as send_telegram:
        xknx.cemi_handler.telegram_received(connect)
        xknx.cemi_handler.telegram_received(ack)
        xknx.cemi_handler.telegram_received(disconnect)
        await asyncio.sleep(0)
        send_telegram.assert_not_called()


async def test_broadcast_message() -> None:
    """Test broadcast message sending."""
    xknx = XKNX()

    test_telegram = Telegram(
        source_address=IndividualAddress("0.0.0"),
        destination_address=GroupAddress("0/0/0"),
        direction=TelegramDirection.OUTGOING,
        tpci=tpci.TDataBroadcast(),
        payload=apci.IndividualAddressRead(),
    )
    with patch("xknx.cemi.CEMIHandler.send_telegram") as send_telegram:
        await xknx.management.send_broadcast(apci.IndividualAddressRead())
        assert send_telegram.call_args_list == [call(test_telegram)]


@pytest.mark.parametrize("rate_limit", [0, 1])
async def test_p2p_rate_limit(
    time_travel: EventLoopClockAdvancer, rate_limit: int
) -> None:
    """Test rate limit for P2P management connections."""
    xknx = XKNX()
    xknx.cemi_handler = AsyncMock()
    ia = IndividualAddress("4.0.1")

    def send_responses(index: int) -> None:
        ack = Telegram(
            source_address=ia,
            destination_address=IndividualAddress(0),
            direction=TelegramDirection.INCOMING,
            tpci=tpci.TAck(index),
        )
        device_desc_resp = Telegram(
            source_address=ia,
            destination_address=IndividualAddress(0),
            direction=TelegramDirection.INCOMING,
            tpci=tpci.TDataConnected(index),
            payload=apci.DeviceDescriptorResponse(),
        )

        xknx.management.process(ack)
        xknx.management.process(device_desc_resp)

    conn = await xknx.management.connect(ia, rate_limit)

    # create task and request data
    task = asyncio.create_task(
        conn.request(
            payload=apci.DeviceDescriptorRead(descriptor=0),
            expected=apci.DeviceDescriptorResponse,
        )
    )

    await asyncio.sleep(0)
    send_responses(0)

    await task

    xknx.cemi_handler.reset_mock()

    # create second task
    task = asyncio.create_task(
        conn.request(
            payload=apci.DeviceDescriptorRead(descriptor=0),
            expected=apci.DeviceDescriptorResponse,
        )
    )
    await asyncio.sleep(0)

    if rate_limit:
        await time_travel(0.5 / rate_limit)

        # the request is still queued
        assert not xknx.cemi_handler.send_telegram.call_args_list

        await time_travel(0.5 / rate_limit)

        # the requests should be sent now, the behaviour should match no rate limit

    assert xknx.cemi_handler.send_telegram.call_args_list == [
        call(
            Telegram(
                destination_address=ia,
                tpci=tpci.TDataConnected(1),
                payload=apci.DeviceDescriptorRead(descriptor=0),
            )
        ),
    ]

    send_responses(1)

    await task



"""
Device is the base class for all implemented devices (e.g. Lights/Switches/Sensors).

It provides basis functionality for reading the state from the KNX bus.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
import logging
from typing import TYPE_CHECKING, Any

from xknx.remote_value import RemoteValue
from xknx.telegram import Telegram
from xknx.telegram.address import DeviceGroupAddress
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite
from xknx.typing import DeviceCallbackType, Self

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class Device(ABC):
    """Base class for devices."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        device_updated_cb: DeviceCallbackType[Self] | None = None,
    ) -> None:
        """Initialize Device class."""
        self.xknx = xknx
        self.name = name
        self.device_updated_cbs: list[DeviceCallbackType[Self]] = []
        if device_updated_cb is not None:
            self.register_device_updated_cb(device_updated_cb)

    def register_state_updater(self) -> None:
        """Register device addresses for StateUpdater."""
        for remote_value in self._iter_remote_values():
            remote_value.register_state_updater()

    def unregister_state_updater(self) -> None:
        """Unregister device addresses from StateUpdater."""
        for remote_value in self._iter_remote_values():
            remote_value.unregister_state_updater()

    def async_start_tasks(self) -> None:
        """Start async background tasks of device."""
        return

    def async_remove_tasks(self) -> None:
        """Remove all tasks of device."""
        return

    @abstractmethod
    def _iter_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes."""
        # yield self.remote_value
        # yield from (<list all used RemoteValue instances>)
        yield from ()

    def group_addresses(self) -> set[DeviceGroupAddress]:
        """Return all group addresses of this Device."""
        return {ga for rv in self._iter_remote_values() for ga in rv.group_addresses()}

    def register_device_updated_cb(
        self, device_updated_cb: DeviceCallbackType[Self]
    ) -> None:
        """Register device updated callback."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(
        self, device_updated_cb: DeviceCallbackType[Self]
    ) -> None:
        """Unregister device updated callback."""
        if device_updated_cb in self.device_updated_cbs:
            self.device_updated_cbs.remove(device_updated_cb)

    def after_update(
        self: Self,
        *args: Any,  # a single argument may be passed if used as a RemoteValue callback
    ) -> None:
        """Execute callbacks after internal state has been changed."""
        for device_callback in self.device_updated_cbs:
            try:
                device_callback(self)
            except Exception:  # pylint: disable=broad-except
                logger.exception(
                    "Unexpected error while processing device_updated_cb for %s",
                    self,
                )

    async def sync(self, wait_for_result: bool = False) -> None:
        """Read states of device from KNX bus."""
        for remote_value in self._iter_remote_values():
            await remote_value.read_state(wait_for_result=wait_for_result)

    def process(self, telegram: Telegram) -> None:
        """Process incoming telegram."""
        if isinstance(telegram.payload, GroupValueWrite):
            self.process_group_write(telegram)
        elif isinstance(telegram.payload, GroupValueResponse):
            self.process_group_response(telegram)
        elif isinstance(telegram.payload, GroupValueRead):
            self.process_group_read(telegram)

    def process_group_read(self, telegram: Telegram) -> None:
        """Process incoming GroupValueRead telegrams."""
        # The default is, that devices don't answer to group reads
        return

    def process_group_response(self, telegram: Telegram) -> None:
        """Process incoming GroupValueResponse telegrams."""
        # Per default mapped to group write.
        self.process_group_write(telegram)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming GroupValueWrite telegrams."""
        # The default is, that devices don't process group writes
        return

    def get_name(self) -> str:
        """Return name of device."""
        return self.name

    def has_group_address(self, group_address: DeviceGroupAddress) -> bool:
        """Test if device has given group address."""
        return any(
            group_address in remote_value.group_addresses()
            for remote_value in self._iter_remote_values()
        )

    def __eq__(self, other: object) -> bool:
        """Compare for quality."""
        return self.__dict__ == other.__dict__



"""
Module for managing setpoint shifting.

DPT 6.010.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from xknx.dpt import DPTArray, DPTBinary, DPTTemperature, DPTValue1Count
from xknx.exceptions import ConversionError, CouldNotParseTelegram

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class SetpointShiftMode(Enum):
    """Enum for setting the setpoint shift mode."""

    DPT6010 = DPTValue1Count
    DPT9002 = DPTTemperature


class RemoteValueSetpointShift(RemoteValue[float]):
    """Abstraction for remote value of KNX DPT 6.010."""

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        after_update_cb: RVCallbackType[float] | None = None,
        setpoint_shift_mode: SetpointShiftMode | None = None,
        setpoint_shift_step: float = 0.1,
    ) -> None:
        """Initialize RemoteValueSetpointShift class."""
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name="Setpoint shift value",
            after_update_cb=after_update_cb,
        )

        self._internal_dpt_class: type[DPTValue1Count | DPTTemperature] | None = (
            setpoint_shift_mode.value if setpoint_shift_mode is not None else None
        )
        self.setpoint_shift_step = setpoint_shift_step

    def to_knx(self, value: float) -> DPTArray:
        """Convert value to payload."""
        if self._internal_dpt_class is None:
            raise ConversionError(
                f"Setpoint shift DPT not initialized for {self.device_name}"
            )
        if self._internal_dpt_class == DPTValue1Count:
            converted_value = int(value / self.setpoint_shift_step)
            return DPTValue1Count.to_knx(converted_value)
        return DPTTemperature.to_knx(value)

    def from_knx(self, payload: DPTArray | DPTBinary) -> float:
        """Convert current payload to value."""
        if self._internal_dpt_class is None:
            self._internal_dpt_class = self._determine_dpt_class(payload)

        payload_value = self._internal_dpt_class.from_knx(payload)
        if self._internal_dpt_class == DPTValue1Count:
            return payload_value * self.setpoint_shift_step
        return payload_value

    def _determine_dpt_class(
        self, payload: DPTArray | DPTBinary
    ) -> type[DPTValue1Count | DPTTemperature]:
        """Test if telegram payload may be parsed."""
        if isinstance(payload, DPTArray):
            payload_length = len(payload.value)
            if payload_length == DPTTemperature.payload_length:
                return DPTTemperature
            if payload_length == DPTValue1Count.payload_length:
                return DPTValue1Count
        raise CouldNotParseTelegram("Payload invalid", payload=str(payload))
